from __future__ import annotations

import itertools
import re
from collections import defaultdict
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import Literal, overload

from xdsl.context import Context
from xdsl.dialect_interfaces import OpAsmDialectInterface
from xdsl.dialects.builtin import DictionaryAttr, ModuleOp
from xdsl.ir import (
    Attribute,
    Block,
    Operation,
    Region,
    SSAValue,
)
from xdsl.irdl import IRDLOperation
from xdsl.utils.exceptions import MultipleSpansParseError
from xdsl.utils.lexer import Input, Span
from xdsl.utils.mlir_lexer import MLIRLexer, MLIRTokenKind

from .attribute_parser import AttrParser  # noqa: TID251
from .generic_parser import ParserState, Position  # noqa: TID251


@dataclass(eq=False)
class ForwardDeclaredValue(SSAValue):
    """
    An SSA value that is used before it is defined.
    It will be replaced to an operation result or a block argument when it is defined.
    """

    @property
    def owner(self) -> Operation | Block:
        raise ValueError("Forward declared values do not have an owner")


@dataclass
class UnresolvedOperand:
    """
    An operand that is not yet resolved in an operation parser.
    It will either be resolved to an SSA value, or to a forward reference of
    an SSA value.
    To resolve it, you need to provide its type.
    """

    span: Span
    """
    The parsing location of the operand name, including the `%`,
    but excluding the optional tuple index.
    """

    index: int
    """The value tuple index, if it is a tuple value."""

    @property
    def operand_name(self) -> str:
        return self.span.text[1:]


class Parser(AttrParser):
    """
    Basic recursive descent parser.

    methods marked try_... will attempt to parse, and return None if they failed.
    If they return None they must make sure to restore all state.

    methods marked parse_... will do "greedy" parsing, meaning they consume
    as much as they can. They will also throw an error if the think they should
    still be parsing. e.g. when parsing a list of numbers separated by '::',
    the following input will trigger an exception:
        1::2::
    Due to the '::' present after the last element. This is useful for parsing lists,
    as a trailing separator is usually considered a syntax error there.

    must_ type parsers are preferred because they are explicit about their failure modes.
    """

    ssa_values: dict[str, tuple[SSAValue, ...]]
    blocks: dict[str, tuple[Block, Span | None]]
    forward_block_references: dict[str, list[Span]]
    """
    Blocks we encountered references to before the definition (must be empty after
    parsing of region completes)
    """
    forward_ssa_references: dict[str, dict[int, ForwardDeclaredValue]]
    """
    SSA values that are referenced, but are not yet defined.
    This field map a name and a tuple index to the forward declared SSA value.
    """

    def __init__(
        self,
        ctx: Context,
        input: str,
        name: str = "<unknown>",
    ) -> None:
        super().__init__(ParserState(MLIRLexer(Input(input, name))), ctx)
        self.ssa_values = dict()
        self.blocks = dict()
        self.forward_block_references = defaultdict(list)
        self.forward_ssa_references = dict()

    def parse_module(self, allow_implicit_module: bool = True) -> ModuleOp:
        module_op: Operation

        if not allow_implicit_module:
            parsed_op = self.parse_optional_operation()

            if parsed_op is None:
                self.raise_error("Could not parse entire input!")

            if not isinstance(parsed_op, ModuleOp):
                self._resume_from(0)
                self.raise_error("builtin.module operation expected", 0)

            module_op = parsed_op
        else:
            parsed_ops: list[Operation] = []

            while self._current_token.kind != MLIRTokenKind.EOF:
                if self._current_token.kind in (
                    MLIRTokenKind.HASH_IDENT,
                    MLIRTokenKind.EXCLAMATION_IDENT,
                ):
                    self._parse_alias_def()
                    continue
                if self._current_token.kind in (MLIRTokenKind.FILE_METADATA_BEGIN,):
                    self._parse_file_metadata_dictionary()
                    continue
                if (parsed_op := self.parse_optional_operation()) is not None:
                    parsed_ops.append(parsed_op)
                    continue
                self.raise_error("Could not parse entire input!")

            if len(parsed_ops) == 0:
                self.raise_error("Could not parse entire input!")

            module_op = (
                parsed_ops[0]
                if isinstance(parsed_ops[0], ModuleOp) and len(parsed_ops) == 1
                else ModuleOp(parsed_ops)
            )

        if self.forward_ssa_references:
            value_names = ", ".join(
                "%" + name for name in self.forward_ssa_references.keys()
            )
            self.raise_error(f"values used but not defined: [{value_names}]")

        return module_op

    def _parse_alias_def(self):
        """
        Parse an attribute or type alias definition with format:
            alias-def           ::= type-alias-def | attribute-alias-def
            type-alias-def      ::= `!` bare-id `=` type
            attribute-alias-def ::= `#` `alias` bare-id `=` attribute
        """
        if (
            token := self._parse_optional_token_in(
                [MLIRTokenKind.EXCLAMATION_IDENT, MLIRTokenKind.HASH_IDENT]
            )
        ) is None:
            self.raise_error("expected attribute name")

        type_or_attr_name = token.text
        if type_or_attr_name in self.attribute_aliases:
            self.raise_error(f"re-declaration of alias '{type_or_attr_name}'")

        self.parse_punctuation("=", "after attribute alias name")
        value = self.parse_attribute()
        self.attribute_aliases[type_or_attr_name] = value

    def _get_block_from_name(self, block_name: Span) -> Block:
        """
        This function takes a span containing a block id (like `^42`) and returns a block.

        If the block definition was not seen yet, we create a forward declaration.
        """
        name = block_name.text[1:]
        if name not in self.blocks:
            self.forward_block_references[name].append(block_name)
            self.blocks[name] = (Block(), None)
        return self.blocks[name][0]

    def _parse_optional_block_arg_list(self, block: Block):
        """
        Parse a block argument list, if present, and add them to the block.

            value-id-and-type-list ::= value-id-and-type (`,` ssa-id-and-type)*
            block-arg-list ::= `(` value-id-and-type-list? `)`
        """
        if self._current_token.kind != MLIRTokenKind.L_PAREN:
            return None

        def parse_argument() -> None:
            """Parse a single block argument with its type."""
            arg_name = self._parse_token(
                MLIRTokenKind.PERCENT_IDENT, "block argument expected"
            ).span
            self.parse_punctuation(":")
            arg_type = self.parse_attribute()
            self.parse_optional_location()

            # Insert the block argument in the block, and register it in the parser
            block_arg = block.insert_arg(arg_type, len(block.args))
            self._register_ssa_definition(arg_name.text[1:], (block_arg,), arg_name)

        self.parse_comma_separated_list(self.Delimiter.PAREN, parse_argument)
        return block

    def _parse_block_body(self, block: Block):
        """
        Parse a block body, which consist of a list of operations.
        The operations are added at the end of the block.
        """
        while (op := self.parse_optional_operation()) is not None:
            block.add_op(op)

    def _parse_block(self) -> Block:
        """
        Parse a block with the following format:
          block ::= block-label operation*
          block-label    ::= block-id block-arg-list? `:`
          block-id       ::= caret-id
          block-arg-list ::= `(` ssa-id-and-type-list? `)`
        """
        name_token = self._parse_token(
            MLIRTokenKind.CARET_IDENT, " in block definition"
        )

        name = name_token.text[1:]
        if name not in self.blocks:
            block = Block()
            self.blocks[name] = (block, name_token.span)
        else:
            block, original_definition = self.blocks[name]
            if original_definition is not None:
                raise MultipleSpansParseError(
                    name_token.span,
                    f"re-declaration of block '{name}'",
                    "originally declared here:",
                    [(original_definition, None)],
                )
            self.forward_block_references.pop(name)

        self._parse_optional_block_arg_list(block)
        self.parse_punctuation(":")
        self._parse_block_body(block)
        return block

    _decimal_integer_regex = re.compile(r"[0-9]+")

    def parse_optional_unresolved_operand(self) -> UnresolvedOperand | None:
        """
        Parse an operand with format `%<value-id>(#<int-literal>)?`, if present.
        The operand may be forward declared.
        """
        name_token = self._parse_optional_token(MLIRTokenKind.PERCENT_IDENT)
        if name_token is None:
            return None

        index = 0
        index_token = self._parse_optional_token(MLIRTokenKind.HASH_IDENT)
        if index_token is not None:
            if re.fullmatch(self._decimal_integer_regex, index_token.text[1:]) is None:
                self.raise_error(
                    "Expected integer as SSA value tuple index", index_token.span
                )
            index = int(index_token.text[1:], 10)

        return UnresolvedOperand(name_token.span, index)

    def parse_unresolved_operand(
        self, msg: str = "operand expected"
    ) -> UnresolvedOperand:
        """
        Parse an operand with format `%<value-id>(#<int-literal>)?`.
        The operand may be forward declared.
        """
        return self.expect(self.parse_optional_unresolved_operand, msg)

    def resolve_operand(self, operand: UnresolvedOperand, type: Attribute) -> SSAValue:
        """
        Resolve an unresolved operand.
        If the operand is not yet defined, it creates a forward reference.
        If the operand is already defined, it returns the corresponding SSA value,
        and checks that the type is consistent.
        """
        name = operand.operand_name

        # If the indexed operand is already used as a forward reference, return it
        if (
            name in self.forward_ssa_references
            and operand.index in self.forward_ssa_references[name]
        ):
            return self.forward_ssa_references[name][operand.index]

        # If the operand is not yet defined, create a forward reference
        if name not in self.ssa_values:
            forward_value = ForwardDeclaredValue(type)
            reference_tuple = self.forward_ssa_references.setdefault(name, {})
            reference_tuple[operand.index] = forward_value
            return forward_value

        # If the operand is already defined, check that the tuple index is in range
        tuple_size = len(self.ssa_values[name])
        if operand.index >= tuple_size:
            self.raise_error(
                "SSA value tuple index out of bounds. "
                f"Tuple is of size {tuple_size} but tried to access element {operand.index}.",
                operand.span,
            )

        # Check that the type is consistent
        resolved = self.ssa_values[name][operand.index]
        if resolved.type != type:
            self.raise_error(
                f"operand is used with type {type}, but has been "
                f"previously used or defined with type {resolved.type}",
                operand.span,
            )

        return resolved

    def parse_optional_operand(self) -> SSAValue | None:
        """
        Parse an operand with format `%<value-id>(#<int-literal>)?`, if present.
        """
        unresolved_operand = self.parse_optional_unresolved_operand()
        if unresolved_operand is None:
            return None

        name = unresolved_operand.operand_name
        index = unresolved_operand.index

        if name not in self.ssa_values.keys():
            self.raise_error(
                "SSA value used before assignment", unresolved_operand.span
            )

        tuple_size = len(self.ssa_values[name])
        if index >= tuple_size:
            self.raise_error(
                "SSA value tuple index out of bounds. "
                f"Tuple is of size {tuple_size} but tried to access element {index}.",
                unresolved_operand.span,
            )

        return self.ssa_values[name][index]

    def parse_operand(self, msg: str = "Expected an operand.") -> SSAValue:
        """Parse an operand with format `%<value-id>`."""
        return self.expect(self.parse_optional_operand, msg)

    def _register_ssa_definition(
        self, name: str, values: Sequence[SSAValue], span: Span
    ) -> None:
        """
        Register an SSA definition in the parsing context.
        In the case the value was already used as a forward reference, the forward
        references are replaced by this value.
        """

        # Check for duplicate SSA value names.
        if name in self.ssa_values:
            self.raise_error(f"SSA value %{name} is already defined", span)

        # Register the SSA values in the context
        self.ssa_values[name] = tuple(values)

        tuple_size = len(values)
        # Check for forward references of this value
        if name in self.forward_ssa_references:
            index_references = self.forward_ssa_references[name]
            del self.forward_ssa_references[name]
            if any(index >= tuple_size for index in index_references):
                self.raise_error(
                    f"SSA value %{name} is referenced with an index "
                    f"larger than its size",
                    span,
                )

            # Replace the forward references with the actual SSA value
            for index, value in index_references.items():
                if index >= tuple_size:
                    self.raise_error(
                        f"SSA value tuple %{name} is referenced with index {index}, but "
                        f"has size {tuple_size}",
                        span,
                    )

                result = values[index]
                if value.type != result.type:
                    result_name = f"%{name}"
                    if tuple_size != 1:
                        result_name = f"%{name}#{index}"
                    self.raise_error(
                        f"Result {result_name} is defined with "
                        f"type {result.type}, but used with type {value.type}",
                        span,
                    )
                value.replace_by(result)

        if SSAValue.is_valid_name(name):
            for val in values:
                val.name_hint = name

    @dataclass
    class UnresolvedArgument:
        """
        A block argument parsed from the assembly.
        Arguments should be parsed by `parse_argument` or `parse_optional_argument`.
        """

        name: Span
        """The name as displayed in the assembly."""

        def resolve(self, type: Attribute) -> Parser.Argument:
            return Parser.Argument(self.name, type)

    @dataclass
    class Argument:
        """
        A block argument parsed from the assembly.
        Arguments should be parsed by `parse_argument` or `parse_optional_argument`.
        """

        name: Span
        """The name as displayed in the assembly."""

        type: Attribute
        """The type of the argument, if any."""

    @overload
    def parse_optional_argument(
        self, expect_type: Literal[True] = True
    ) -> Argument | None: ...

    @overload
    def parse_optional_argument(
        self, expect_type: Literal[False]
    ) -> UnresolvedArgument | None: ...

    @overload
    def parse_optional_argument(
        self, expect_type: bool = True
    ) -> UnresolvedArgument | Argument | None: ...

    def parse_optional_argument(
        self, expect_type: bool = True
    ) -> UnresolvedArgument | Argument | None:
        """
        Parse a block argument, if present, with format:
          arg ::= percent-id `:` type
        if `expect_type` is False, the type is not parsed.
        """

        # The argument name
        name_token = self._parse_optional_token(MLIRTokenKind.PERCENT_IDENT)
        if name_token is None:
            return None

        # The argument type
        if expect_type:
            self.parse_punctuation(":", " after block argument name!")
            type = self.parse_type()
            return self.Argument(name_token.span, type)
        else:
            return self.UnresolvedArgument(name_token.span)

    @overload
    def parse_argument(self, *, expect_type: Literal[True] = True) -> Argument: ...

    @overload
    def parse_argument(self, *, expect_type: Literal[False]) -> UnresolvedArgument: ...

    @overload
    def parse_argument(
        self, *, expect_type: bool = True
    ) -> UnresolvedArgument | Argument: ...

    def parse_argument(
        self, *, expect_type: bool = True
    ) -> UnresolvedArgument | Argument:
        """
        Parse a block argument with format:
          arg ::= percent-id `:` type
        if `expect_type` is False, the type is not parsed.
        """

        arg = self.parse_optional_argument(expect_type)
        if arg is None:
            self.raise_error("Expected block argument!")
        return arg

    def parse_optional_region(
        self, arguments: Iterable[Argument] | None = None
    ) -> Region | None:
        """
        Parse a region, if present, with format:
          region ::= `{` entry-block? block* `}`
        If `arguments` is provided, the entry block will use these as block arguments,
        and the entry-block cannot be labeled. It also cannot be empty, unless it is the
        only block in the region.
        """
        # Check if a region is present.
        if self.parse_optional_punctuation("{") is None:
            return None

        region = Region()

        # Create a new scope for values and blocks.
        # Outside blocks cannot be referenced from inside the region, and vice versa.
        # Outside values can be referenced from inside the region, but the region
        # values cannot be referred to from outside the region.
        old_ssa_values = self.ssa_values.copy()
        old_blocks = self.blocks
        old_forward_blocks = self.forward_block_references
        self.blocks = dict()
        self.forward_block_references = defaultdict(list)

        # Parse the entry block without label if arguments are provided.
        # Since the entry block cannot be jumped to, this is fine.
        if arguments is not None:
            # Check that the provided arguments have types.
            arg_types = [arg.type for arg in arguments]

            # Check that the entry block has no label.
            # Since a multi-block region block must have a terminator, there isn't a
            # possibility of having an empty entry block, and thus parsing the label directly.
            if self._current_token.kind == MLIRTokenKind.CARET_IDENT:
                self.raise_error("invalid block name in region with named arguments")

            # Set the block arguments in the context
            entry_block = Block(arg_types=arg_types)
            for block_arg, arg in zip(entry_block.args, arguments):
                self._register_ssa_definition(arg.name.text[1:], (block_arg,), arg.name)

            # Parse the entry block body
            self._parse_block_body(entry_block)
            region.add_block(entry_block)

        # If no arguments was provided, parse the entry block if present.
        elif self._current_token.kind not in (
            MLIRTokenKind.CARET_IDENT,
            MLIRTokenKind.R_BRACE,
        ):
            block = Block()
            self._parse_block_body(block)
            region.add_block(block)

        # Parse the region blocks.
        # In the case where arguments are provided, the entry block is already parsed,
        # and the following blocks will have a label (since the entry block will parse
        # greedily all operations).
        # In the case where no arguments areprovided, the entry block can either have a
        # label or not.
        while self.parse_optional_punctuation("}") is None:
            block = self._parse_block()
            region.add_block(block)

        # Finally, check that all forward block references have been resolved.
        if self.forward_block_references:
            pos = self.pos
            raise MultipleSpansParseError(
                Span(pos, pos + 1, self.lexer.input),
                "region ends with missing block declarations for block(s) {}".format(
                    ", ".join(self.forward_block_references.keys())
                ),
                "dangling block references:",
                [
                    (
                        span,
                        f'reference to block "{span.text}" without implementation',
                    )
                    for span in itertools.chain(*self.forward_block_references.values())
                ],
            )

        # Close the value and block scope.
        self.ssa_values = old_ssa_values
        self.blocks = old_blocks
        self.forward_block_references = old_forward_blocks

        return region

    def parse_region(self, arguments: Iterable[Argument] | None = None) -> Region:
        """
        Parse a region with format:
          region ::= `{` entry-block? block* `}`
        If `arguments` is provided, the entry block will use these as block arguments,
        and the entry-block cannot be labeled. It also cannot be empty, unless it is the
        only block in the region.
        """
        region = self.parse_optional_region(arguments)
        if region is None:
            self.raise_error("Expected region!")
        return region

    def parse_region_list(self) -> list[Region]:
        """
        Parse the list of operation regions.
        If no regions are present, return an empty list.
        Parse a list of regions with format:
           regions-list ::= `(` region (`,` region)* `)`
        """
        if self._current_token.kind == MLIRTokenKind.L_PAREN:
            return self.parse_comma_separated_list(
                self.Delimiter.PAREN, self.parse_region, " in operation region list"
            )
        return []

    def parse_op(self) -> Operation:
        return self.parse_operation()

    def parse_optional_attr_dict_with_keyword(
        self, reserved_attr_names: Iterable[str] = ()
    ) -> DictionaryAttr | None:
        """
        Parse a dictionary attribute, preceeded with `attributes` keyword, if the
        keyword is present.
        This is intended to be used in operation custom assembly format.
        `reserved_attr_names` contains names that should not be present in the attribute
        dictionary, and usually correspond to the names of the attributes that are
        already passed through the operation custom assembly format.
        """
        if self.parse_optional_keyword("attributes") is None:
            return None
        return self.parse_optional_attr_dict_with_reserved_attr_names(
            reserved_attr_names
        )

    def parse_optional_attr_dict_with_reserved_attr_names(
        self, reserved_attr_names: Iterable[str] = ()
    ) -> DictionaryAttr | None:
        """
        Parse a dictionary attribute if present.
        This is intended to be used in operation custom assembly format.
        `reserved_attr_names` contains names that should not be present in the attribute
        dictionary, and usually correspond to the names of the attributes that are
        already passed through the operation custom assembly format.
        """
        begin_pos = self.lexer.pos
        attr = self._parse_builtin_dict_attr()

        for reserved_name in reserved_attr_names:
            if reserved_name in attr.data:
                self.raise_error(
                    f"Attribute dictionary entry '{reserved_name}' is already passed "
                    "through the operation custom assembly format.",
                    Span(begin_pos, begin_pos, self.lexer.input),
                )
        return attr

    def parse_optional_operation(self) -> Operation | None:
        """
        Parse an operation with format:
            operation             ::= op-result-list? (generic-operation | custom-operation)
            generic-operation     ::= string-literal `(` value-use-list? `)`  successor-list?
                                      properties? region-list? dictionary-attribute? `:`
                                      function-type location?
            custom-operation      ::= bare-id custom-operation-format
            op-result-list        ::= op-result (`,` op-result)* `=`
            op-result             ::= value-id (`:` integer-literal)
            successor-list        ::= `[` successor (`,` successor)* `]`
            successor             ::= caret-id (`:` block-arg-list)?
            region-list           ::= `(` region (`,` region)* `)`
            dictionary-attribute  ::= `{` (attribute-entry (`,` attribute-entry)*)? `}`
            properties            ::= `<` dictionary-attribute `>`
        """
        if self._current_token.kind not in (
            MLIRTokenKind.PERCENT_IDENT,
            MLIRTokenKind.BARE_IDENT,
            MLIRTokenKind.STRING_LIT,
        ):
            return None
        return self.parse_operation()

    def parse_operation(self) -> Operation:
        """
        Parse an operation with format:
            operation             ::= op-result-list? (generic-operation | custom-operation)
            generic-operation     ::= string-literal `(` value-use-list? `)`  successor-list?
                                      properties? region-list? dictionary-attribute? `:`
                                      function-type location?
            custom-operation      ::= bare-id custom-operation-format
            op-result-list        ::= op-result (`,` op-result)* `=`
            op-result             ::= value-id (`:` integer-literal)
            successor-list        ::= `[` successor (`,` successor)* `]`
            successor             ::= caret-id (`:` block-arg-list)?
            region-list           ::= `(` region (`,` region)* `)`
            dictionary-attribute  ::= `{` (attribute-entry (`,` attribute-entry)*)? `}`
            properties            ::= `<` dictionary-attribute `>`
        """
        # Parse the operation results
        op_loc = self._current_token.span
        bound_results = self._parse_op_result_list()

        if (
            op_name := self._parse_optional_token(MLIRTokenKind.BARE_IDENT)
        ) is not None:
            # Custom operation format
            op_type = self._get_op_by_name(op_name.text)
            dialect_name = op_type.dialect_name()
            self._parser_state.dialect_stack.append(dialect_name)
            op = op_type.parse(self)
            self._parser_state.dialect_stack.pop()
        else:
            # Generic operation format
            op_name = self.expect(
                self.parse_optional_str_literal, "operation name expected"
            )
            op_type = self._get_op_by_name(op_name)
            dialect_name = op_type.dialect_name()
            self._parser_state.dialect_stack.append(dialect_name)
            op = self._parse_generic_operation(op_type)
            self._parser_state.dialect_stack.pop()

        n_bound_results = sum(r[1] for r in bound_results)
        if (n_bound_results != 0) and (len(op.results) != n_bound_results):
            self.raise_error(
                f"Operation has {len(op.results)} results, "
                f"but was given {n_bound_results} to bind.",
                at_position=op_loc,
            )

        # Register the result SSA value names in the parser
        res_idx = 0
        for res_span, res_size in bound_results:
            ssa_val_name = res_span.text[1:]  # Removing the leading '%'
            self._register_ssa_definition(
                ssa_val_name, op.results[res_idx : res_idx + res_size], res_span
            )
            res_idx += res_size

        return op

    def _get_op_by_name(self, name: str) -> type[Operation]:
        """
        Get an operation type by its name.
        Raises an error if the operation is not registered, and if unregistered
        dialects are not allowed.
        """
        if op_type := self.ctx.get_optional_op(
            name, dialect_stack=self._parser_state.dialect_stack
        ):
            return op_type
        self.raise_error(f"Operation {name} is not registered")

    def _parse_op_result(self) -> tuple[Span, int]:
        """
        Parse an operation result.
        Returns the span of the SSA value name (including the `%`), and the size of the
        value tuple (by default 1).
        """
        value_token = self._parse_token(
            MLIRTokenKind.PERCENT_IDENT, "Expected result SSA value!"
        )
        if self._parse_optional_token(MLIRTokenKind.COLON) is None:
            return (value_token.span, 1)

        size_token = self._parse_token(
            MLIRTokenKind.INTEGER_LIT, "Expected SSA value tuple size"
        )
        size = size_token.kind.get_int_value(size_token.span)
        return (value_token.span, size)

    def _parse_op_result_list(self) -> list[tuple[Span, int]]:
        """
        Parse the list of operation results.
        If no results are present, returns an empty list.
        Each result is a tuple of the span of the SSA value name (including the `%`),
        and the size of the value tuple (by default 1).
        """
        if self._current_token.kind == MLIRTokenKind.PERCENT_IDENT:
            res = self.parse_comma_separated_list(
                self.Delimiter.NONE, self._parse_op_result, " in operation result list"
            )
            self.parse_punctuation("=", " after operation result list")
            return res
        return []

    def parse_optional_attr_dict(self) -> dict[str, Attribute]:
        return self.parse_optional_dictionary_attr_dict()

    def parse_optional_properties_dict(self) -> dict[str, Attribute]:
        """
        Parse a property dictionary, if present, with format:
            dictionary-attribute  ::= `{` (attribute-entry (`,` attribute-entry)*)? `}`
            properties            ::= `<` dictionary-attribute `>`
        """
        if self.parse_optional_punctuation("<") is None:
            return dict()

        entries = self.parse_comma_separated_list(
            self.Delimiter.BRACES, self._parse_attribute_entry
        )
        self.parse_punctuation(">")

        if (key := self._find_duplicated_key(entries)) is not None:
            self.raise_error(f"Duplicate key '{key}' in properties dictionary")

        return dict(entries)

    def resolve_operands(
        self,
        args: Sequence[UnresolvedOperand],
        input_types: Sequence[Attribute],
        error_pos: Position,
    ) -> Sequence[SSAValue]:
        """
        Resolve unresolved operands. For each operand in `args` and its corresponding input
        type the following happens:

        If the operand is not yet defined, it creates a forward reference.
        If the operand is already defined, it returns the corresponding SSA value,
        and checks that the type is consistent.

        If the length of args and input_types does not match, an error is raised at
        the location error_pos.
        """
        length = len(list(input_types))
        if len(args) != length:
            self.raise_error(
                f"expected {length} operand types but had {len(args)}",
                error_pos,
            )

        return [
            self.resolve_operand(operand, type)
            for operand, type in zip(args, input_types)
        ]

    def _parse_generic_operation(self, op_type: type[Operation]) -> Operation:
        """
        Parse a generic operation with format:
            generic-operation     ::= string-literal `(` value-use-list? `)`  successor-list?
                                      properties? region-list? dictionary-attribute? `:`
                                      function-type location?
            custom-operation      ::= bare-id custom-operation-format
            op-result-list        ::= op-result (`,` op-result)* `=`
            op-result             ::= value-id (`:` integer-literal)
            successor-list        ::= `[` successor (`,` successor)* `]`
            successor             ::= caret-id (`:` block-arg-list)?
            region-list           ::= `(` region (`,` region)* `)`
            dictionary-attribute  ::= `{` (attribute-entry (`,` attribute-entry)*)? `}`
            properties            ::= `<` dictionary-attribute `>`
        """
        # Parse arguments
        args = self.parse_op_args_list()

        # Parse successors
        successors = self.parse_optional_successors()
        if successors is None:
            successors = []

        # Parse attribute dictionary
        properties = self.parse_optional_properties_dict()

        # Parse regions
        regions = self.parse_region_list()

        # Parse attribute dictionary
        attributes = self.parse_optional_attr_dict()

        self.parse_punctuation(":", "function type signature expected")

        func_type_pos = self._current_token.span.start

        # Parse function type
        func_type = self.parse_function_type()

        self.parse_optional_location()

        operands = self.resolve_operands(args, func_type.inputs.data, func_type_pos)

        # Properties retrocompatibility :
        # We extract properties from the attribute dictionary by name.
        if issubclass(op_type, IRDLOperation):
            op_def = op_type.get_irdl_definition()
            for property_name in op_def.properties.keys():
                if property_name in attributes and property_name not in properties:
                    properties[property_name] = attributes.pop(property_name)

        return op_type.create(
            operands=operands,
            result_types=func_type.outputs.data,
            properties=properties,
            attributes=attributes,
            successors=successors,
            regions=regions,
        )

    def parse_optional_successor(self) -> Block | None:
        """
        Parse a successor with format:
            successor      ::= caret-id
        """
        block_token = self._parse_optional_token(MLIRTokenKind.CARET_IDENT)
        if block_token is None:
            return None
        name = block_token.text[1:]
        if name not in self.blocks:
            self.forward_block_references[name].append(block_token.span)
            self.blocks[name] = (Block(), None)
        return self.blocks[name][0]

    def parse_successor(self) -> Block:
        """
        Parse a successor with format:
            successor      ::= caret-id
        """
        return self.expect(self.parse_optional_successor, "successor expected")

    def parse_optional_successors(self) -> list[Block] | None:
        """
        Parse a list of successors, if present, with format
            successor-list ::= `[` successor (`,` successor)* `]`
            successor      ::= caret-id
        """
        if self._current_token.kind != MLIRTokenKind.L_SQUARE:
            return None
        return self.parse_successors()

    def parse_successors(self) -> list[Block]:
        """
        Parse a list of successors with format:
            successor-list ::= `[` successor (`,` successor)* `]`
            successor      ::= caret-id
        """
        return self.parse_comma_separated_list(
            self.Delimiter.SQUARE,
            lambda: self.expect(self.parse_successor, "block-id expected"),
        )

    def parse_op_args_list(self) -> list[UnresolvedOperand]:
        """
        Parse a list of arguments with format:
           args-list ::= `(` value-use-list? `)`
           value-use-list ::= `%` suffix-id (`,` `%` suffix-id)*
        """
        return self.parse_comma_separated_list(
            self.Delimiter.PAREN,
            self.parse_unresolved_operand,
            " in operation argument list",
        )

    def _parse_resource(
        self, dialect_name: str, interface: OpAsmDialectInterface
    ) -> None:
        key = self._parse_dialect_resource_handle(dialect_name, interface)
        self._parse_token(MLIRTokenKind.COLON, "expected `:`")
        value = self.parse_str_literal()

        try:
            interface.parse_resource(key, value)
        except Exception as e:
            self.raise_error(f"got an error when parsing a resource: {e}")

    def _parse_single_dialect_resources(self) -> None:
        dialect_name = self._parse_token(
            MLIRTokenKind.BARE_IDENT, "Expected a dialect name"
        )
        self._parse_token(MLIRTokenKind.COLON, "expected `:`")

        dialect = self.ctx.get_optional_dialect(dialect_name.text)
        if dialect is None:
            self.raise_error(f"dialect {dialect_name.text} is not registered")

        interface = dialect.get_interface(OpAsmDialectInterface)
        if not interface:
            self.raise_error(
                f"dialect {dialect.name} doesn't have an OpAsmDialectInterface interface"
            )

        self.parse_comma_separated_list(
            self.Delimiter.BRACES, lambda: self._parse_resource(dialect.name, interface)
        )

    def _parse_dialect_resources(self) -> None:
        self.parse_comma_separated_list(
            self.Delimiter.BRACES, self._parse_single_dialect_resources
        )

    def _parse_external_resources(self) -> None:
        raise NotImplementedError("Currently only dialect resources are supported")

    def _parse_metadata_element(self) -> None:
        resource_type = self._parse_token(
            MLIRTokenKind.BARE_IDENT, "Expected a resource type key"
        )

        self._parse_token(MLIRTokenKind.COLON, "expected `:`")

        match resource_type.text:
            case "dialect_resources":
                self._parse_dialect_resources()
            case "external_resources":
                self._parse_external_resources()
            case _:
                self.raise_error(
                    f"got an unexpected key in file metadata: {resource_type.text}"
                )

    def _parse_file_metadata_dictionary(self) -> None:
        """
        Parse metadata section {-# ... #-} of the file.
        Returns None since results are stored in the context object.
        """
        self.parse_comma_separated_list(
            self.Delimiter.METADATA_TOKEN, self._parse_metadata_element
        )



from xdsl.backend.x86.register_stack import X86RegisterStack
from xdsl.dialects import x86


def test_default_reserved_registers():
    register_stack = X86RegisterStack.get()

    for reg in (
        x86.register.RAX,
        x86.register.RDX,
        x86.register.RSP,
    ):
        available_before = register_stack.available_registers.copy()
        register_stack.push(reg)
        assert available_before == register_stack.available_registers


def test_push_infinite_register():
    register_stack = X86RegisterStack()

    infinite0 = x86.AVX2RegisterType.infinite_register(0)
    register_stack.push(infinite0)
    assert register_stack.pop(x86.AVX2RegisterType) == infinite0


def test_push_register():
    register_stack = X86RegisterStack.get()

    register_stack.push(x86.register.YMM0)
    assert register_stack.pop(x86.AVX2RegisterType) == x86.register.YMM0

    register_stack.push(x86.register.RAX)
    assert register_stack.pop(x86.GeneralRegisterType) == x86.register.RAX



import pytest

from xdsl.transforms.memref_stream_interleave import IndexAndFactor, factors


@pytest.mark.parametrize(
    "num,expected_factors",
    [
        (-1, ()),
        (0, ()),
        (1, (1,)),
        (2, (1, 2)),
        (3, (1, 3)),
        (4, (1, 2, 4)),
        (6, (1, 2, 3, 6)),
        (24, (1, 2, 3, 4, 6, 8, 12, 24)),
    ],
)
def test_factors_parametrized(num: int, expected_factors: tuple[int, ...]):
    assert factors(num) == expected_factors


@pytest.mark.parametrize(
    "indices_and_factors,expected_res",
    [
        ((), None),
        (((0, 1),), (0, 1)),
        (((0, 1), (1, 1)), (1, 1)),
        (((0, 1), (1, 2), (1, 3)), (1, 3)),
        (((0, 1), (1, 4), (1, 5)), (1, 5)),
        (((0, 1), (0, 3), (1, 4), (1, 11), (1, 44)), (1, 4)),
    ],
)
def test_choose_index_and_factor(
    indices_and_factors: tuple[tuple[int, int], ...],
    expected_res: tuple[int, int] | None,
):
    named_tuples = tuple(
        IndexAndFactor(index, factor) for index, factor in indices_and_factors
    )
    assert IndexAndFactor.choose(named_tuples, 4) == expected_res



import sys
from collections.abc import Callable

from xdsl.ir import Dialect
from xdsl.utils.dialect_loader import IRDLDialectFinder


def get_all_dialects() -> dict[str, Callable[[], Dialect]]:
    """Returns all available dialects."""

    def get_accfg():
        from xdsl.dialects.accfg import ACCFG

        return ACCFG

    def get_affine():
        from xdsl.dialects.affine import Affine

        return Affine

    def get_air():
        from xdsl.dialects.experimental.air import AIR

        return AIR

    def get_arith():
        from xdsl.dialects.arith import Arith

        return Arith

    def get_arm():
        from xdsl.dialects.arm import ARM

        return ARM

    def get_arm_func():
        from xdsl.dialects.arm_func import ARM_FUNC

        return ARM_FUNC

    def get_arm_neon():
        from xdsl.dialects.arm_neon import ARM_NEON

        return ARM_NEON

    def get_bigint():
        from xdsl.dialects.bigint import BigInt

        return BigInt

    def get_bufferization():
        from xdsl.dialects.bufferization import Bufferization

        return Bufferization

    def get_builtin():
        from xdsl.dialects.builtin import Builtin

        return Builtin

    def get_cf():
        from xdsl.dialects.cf import Cf

        return Cf

    def get_cmath():
        from xdsl.dialects.cmath import Cmath

        return Cmath

    def get_comb():
        from xdsl.dialects.comb import Comb

        return Comb

    def get_csl():
        from xdsl.dialects.csl import CSL

        return CSL

    def get_csl_stencil():
        from xdsl.dialects.csl.csl_stencil import CSL_STENCIL

        return CSL_STENCIL

    def get_csl_wrapper():
        from xdsl.dialects.csl.csl_wrapper import CSL_WRAPPER

        return CSL_WRAPPER

    def get_dlti():
        from xdsl.dialects.dlti import DLTI

        return DLTI

    def get_dmp():
        from xdsl.dialects.experimental.dmp import DMP

        return DMP

    def get_emitc():
        from xdsl.dialects.emitc import EmitC

        return EmitC

    def get_eqsat():
        from xdsl.dialects.eqsat import EqSat

        return EqSat

    def get_fir():
        from xdsl.dialects.experimental.fir import FIR

        return FIR

    def get_fsm():
        from xdsl.dialects.fsm import FSM

        return FSM

    def get_func():
        from xdsl.dialects.func import Func

        return Func

    def get_gpu():
        from xdsl.dialects.gpu import GPU

        return GPU

    def get_hlfir():
        from xdsl.dialects.experimental.hlfir import HLFIR

        return HLFIR

    def get_hls():
        from xdsl.dialects.experimental.hls import HLS

        return HLS

    def get_hw():
        from xdsl.dialects.hw import HW

        return HW

    def get_linalg():
        from xdsl.dialects.linalg import Linalg

        return Linalg

    def get_irdl():
        from xdsl.dialects.irdl.irdl import IRDL

        return IRDL

    def get_llvm():
        from xdsl.dialects.llvm import LLVM

        return LLVM

    def get_ltl():
        from xdsl.dialects.ltl import LTL

        return LTL

    def get_math():
        from xdsl.dialects.math import Math

        return Math

    def get_math_xdsl():
        from xdsl.dialects.math_xdsl import MathXDSL

        return MathXDSL

    def get_memref():
        from xdsl.dialects.memref import MemRef

        return MemRef

    def get_memref_stream():
        from xdsl.dialects.memref_stream import MemRefStream

        return MemRefStream

    def get_ml_program():
        from xdsl.dialects.ml_program import MLProgram

        return MLProgram

    def get_mod_arith():
        from xdsl.dialects.mod_arith import ModArith

        return ModArith

    def get_mpi():
        from xdsl.dialects.mpi import MPI

        return MPI

    def get_omp():
        from xdsl.dialects.omp import OMP

        return OMP

    def get_pdl():
        from xdsl.dialects.pdl import PDL

        return PDL

    def get_pdl_interp():
        from xdsl.dialects.pdl_interp import PDLInterp

        return PDLInterp

    def get_printf():
        from xdsl.dialects.printf import Printf

        return Printf

    def get_ptr_xdsl():
        from xdsl.dialects.ptr import Ptr

        return Ptr

    def get_riscv_debug():
        from xdsl.dialects.riscv_debug import RISCV_Debug

        return RISCV_Debug

    def get_riscv():
        from xdsl.dialects.riscv import RISCV

        return RISCV

    def get_riscv_func():
        from xdsl.dialects.riscv_func import RISCV_Func

        return RISCV_Func

    def get_riscv_scf():
        from xdsl.dialects.riscv_scf import RISCV_Scf

        return RISCV_Scf

    def get_riscv_cf():
        from xdsl.dialects.riscv_cf import RISCV_Cf

        return RISCV_Cf

    def get_riscv_snitch():
        from xdsl.dialects.riscv_snitch import RISCV_Snitch

        return RISCV_Snitch

    def get_scf():
        from xdsl.dialects.scf import Scf

        return Scf

    def get_seq():
        from xdsl.dialects.seq import Seq

        return Seq

    def get_smt():
        from xdsl.dialects.smt import SMT

        return SMT

    def get_snitch():
        from xdsl.dialects.snitch import Snitch

        return Snitch

    def get_snitch_runtime():
        from xdsl.dialects.snitch_runtime import SnitchRuntime

        return SnitchRuntime

    def get_snitch_stream():
        from xdsl.dialects.snitch_stream import SnitchStream

        return SnitchStream

    def get_stablehlo():
        from xdsl.dialects.stablehlo import StableHLO

        return StableHLO

    def get_stencil():
        from xdsl.dialects.stencil import Stencil

        return Stencil

    def get_stim():
        from xdsl.dialects.stim import Stim

        return Stim

    def get_symref():
        from xdsl.dialects.symref import Symref

        return Symref

    def get_tensor():
        from xdsl.dialects.tensor import Tensor

        return Tensor

    def get_test():
        from xdsl.dialects.test import Test

        return Test

    def get_tosa():
        from xdsl.dialects.tosa import TOSA

        return TOSA

    def get_varith():
        from xdsl.dialects.varith import Varith

        return Varith

    def get_vector():
        from xdsl.dialects.vector import Vector

        return Vector

    def get_wasm():
        from xdsl.dialects.wasm import Wasm

        return Wasm

    def get_x86():
        from xdsl.dialects.x86 import X86

        return X86

    def get_x86_func():
        from xdsl.dialects.x86_func import X86_FUNC

        return X86_FUNC

    def get_transform():
        from xdsl.dialects.transform import Transform

        return Transform

    return {
        "accfg": get_accfg,
        "affine": get_affine,
        "air": get_air,
        "arith": get_arith,
        "arm": get_arm,
        "arm_func": get_arm_func,
        "arm_neon": get_arm_neon,
        "bigint": get_bigint,
        "bufferization": get_bufferization,
        "builtin": get_builtin,
        "cf": get_cf,
        "cmath": get_cmath,
        "comb": get_comb,
        "csl": get_csl,
        "csl_stencil": get_csl_stencil,
        "csl_wrapper": get_csl_wrapper,
        "dlti": get_dlti,
        "dmp": get_dmp,
        "emitc": get_emitc,
        "eqsat": get_eqsat,
        "fir": get_fir,
        "fsm": get_fsm,
        "func": get_func,
        "gpu": get_gpu,
        "hlfir": get_hlfir,
        "hls": get_hls,
        "hw": get_hw,
        "linalg": get_linalg,
        "irdl": get_irdl,
        "llvm": get_llvm,
        "ltl": get_ltl,
        "math": get_math,
        "math_xdsl": get_math_xdsl,
        "memref": get_memref,
        "memref_stream": get_memref_stream,
        "ml_program": get_ml_program,
        "mod_arith": get_mod_arith,
        "mpi": get_mpi,
        "omp": get_omp,
        "pdl": get_pdl,
        "pdl_interp": get_pdl_interp,
        "printf": get_printf,
        "ptr_xdsl": get_ptr_xdsl,
        "riscv": get_riscv,
        "riscv_debug": get_riscv_debug,
        "riscv_func": get_riscv_func,
        "riscv_scf": get_riscv_scf,
        "riscv_cf": get_riscv_cf,
        "riscv_snitch": get_riscv_snitch,
        "scf": get_scf,
        "seq": get_seq,
        "smt": get_smt,
        "snitch": get_snitch,
        "snrt": get_snitch_runtime,
        "snitch_stream": get_snitch_stream,
        "stablehlo": get_stablehlo,
        "stencil": get_stencil,
        "stim": get_stim,
        "symref": get_symref,
        "tensor": get_tensor,
        "test": get_test,
        "tosa": get_tosa,
        "varith": get_varith,
        "vector": get_vector,
        "wasm": get_wasm,
        "x86": get_x86,
        "x86_func": get_x86_func,
        "transform": get_transform,
    }


# Add the IRDLDialectFinder to the meta path as last resort, i.e, it will look for a
# .irdl implementation if no .py implementation is found.
sys.meta_path.append(IRDLDialectFinder(get_all_dialects))



from typing import Any, ClassVar, get_origin

from xdsl.utils.exceptions import PyRDLError


def is_classvar(annotation: Any) -> bool:
    """
    The type annotation can be one of
     * `ClassVar[MyType]`,
     * `ClassVar`, or
     * `"ClassVar[MyType]"`.
    """
    return (
        get_origin(annotation) is ClassVar
        or annotation is ClassVar
        or (isinstance(annotation, str) and annotation.startswith("ClassVar"))
    )


def is_const_classvar(
    field_name: str, annotation: Any, error_type: type[PyRDLError]
) -> bool:
    """
    Operation definitions may only have `*_def` fields or constant class variables,
    where the constness is defined by convention with an UPPER_CASE name and enforced by
    pyright.
    The type annotation can be one of
     * `ClassVar[MyType]`,
     * `ClassVar`, or
     * `"ClassVar[MyType]"`.
    This function throws an exception on a non UPPER_CASE class variable.
    """
    if not is_classvar(annotation):
        return False
    if not field_name.isupper():
        raise error_type(f'Invalid ClassVar name "{field_name}", must be uppercase.')
    return True



from xdsl.builder import Builder, ImplicitBuilder
from xdsl.dialects import arith, func
from xdsl.dialects.builtin import ModuleOp, StringAttr, SymbolRefAttr, i32
from xdsl.interpreter import Interpreter
from xdsl.interpreters.arith import ArithFunctions
from xdsl.interpreters.func import FuncFunctions


@ModuleOp
@Builder.implicit_region
def outer_module():
    @ModuleOp
    @Builder.implicit_region
    def nested():
        with ImplicitBuilder(func.FuncOp("nested_func", ((), (i32,))).body):
            a = arith.ConstantOp.from_int_and_width(42, 32).result
            func.ReturnOp(a)

    nested.sym_name = StringAttr("nested")
    _call_op = func.CallOp(
        SymbolRefAttr("nested", ("nested_func",)),
        (),
        (i32,),
    )


def test_nested_symbol_lookup():
    interpreter = Interpreter(outer_module)
    interpreter.register_implementations(FuncFunctions())
    interpreter.register_implementations(ArithFunctions())
    assert interpreter.call_op(SymbolRefAttr("nested", ("nested_func",)), ()) == (42,)
    call_op = outer_module.regions[0].block.last_op
    assert isinstance(call_op, func.CallOp)
    assert interpreter.run_op(call_op, ()) == (42,)



#!/usr/bin/env python3
"""Benchmarks for the pattern rewriter of the xDSL implementation."""

from typing import cast

from benchmarks.workloads import WorkloadBuilder
from xdsl.context import Context
from xdsl.dialects.arith import (
    AddiOp,
    Arith,
    ConstantOp,
    SubiOp,
)
from xdsl.dialects.builtin import Builtin, IntegerAttr, IntegerType, ModuleOp
from xdsl.ir import Region
from xdsl.ir.post_order import PostOrderIterator
from xdsl.irdl import VarIRConstruct, get_variadic_sizes
from xdsl.parser import Parser as XdslParser
from xdsl.pattern_rewriter import PatternRewriter, Worklist
from xdsl.rewriter import InsertPoint
from xdsl.traits import (
    HasCanonicalizationPatternsTrait,
    MemoryEffect,
)
from xdsl.transforms.canonicalization_patterns.utils import const_evaluate_operand
from xdsl.transforms.canonicalize import CanonicalizePass
from xdsl.transforms.dead_code_elimination import (
    LiveSet,
    is_trivially_dead,
    result_only_effects,
    would_be_trivially_dead,
)
from xdsl.transforms.test_constant_folding import (
    TestConstantFoldingPass,
    TestSpecialisedConstantFoldingPass,
)

CTX = Context(allow_unregistered=True)
CTX.load_dialect(Arith)
CTX.load_dialect(Builtin)

CANONICALIZE_PASS = CanonicalizePass()
CONSTANT_FOLDING_SIMPLE_PASS = TestConstantFoldingPass()
CONSTANT_FOLDING_SPECIALISED_PASS = TestSpecialisedConstantFoldingPass()


def parse_module(context: Context, contents: str) -> ModuleOp:
    """Parse a MLIR file as a module."""
    parser = XdslParser(context, contents)
    return parser.parse_module()


class ConstantFolding:
    """Benchmark rewriting in xDSL."""

    WORKLOAD_CONSTANT_20 = parse_module(CTX, WorkloadBuilder.constant_folding(20))
    WORKLOAD_CONSTANT_100 = parse_module(CTX, WorkloadBuilder.constant_folding(100))
    WORKLOAD_CONSTANT_1000 = parse_module(CTX, WorkloadBuilder.constant_folding(1_000))

    workload_constant_20: ModuleOp
    workload_constant_100: ModuleOp
    workload_constant_1000: ModuleOp

    def setup(self) -> None:
        """Setup the benchmarks."""
        self.setup_constant_folding_20()
        self.setup_constant_folding_100()
        self.setup_constant_folding_1000()

    def setup_constant_folding_20(self) -> None:
        """Setup the constant folding 20 items benchmark."""
        self.workload_constant_20 = ConstantFolding.WORKLOAD_CONSTANT_20.clone()

    def time_constant_folding_20(self) -> None:
        """Time canonicalizing constant folding for 20 items."""
        CANONICALIZE_PASS.apply(CTX, self.workload_constant_20)

    def time_constant_folding_simple_20(self) -> None:
        """Time simple constant folding for 20 items."""
        CONSTANT_FOLDING_SIMPLE_PASS.apply(CTX, self.workload_constant_20)

    def time_constant_folding_specialised_20(self) -> None:
        """Time simple constant folding for 20 items."""
        CONSTANT_FOLDING_SPECIALISED_PASS.apply(CTX, self.workload_constant_20)

    def setup_constant_folding_100(self) -> None:
        """Setup the constant folding 100 items benchmark."""
        self.workload_constant_100 = ConstantFolding.WORKLOAD_CONSTANT_100.clone()

    def time_constant_folding_100(self) -> None:
        """Time canonicalizing constant folding for 100 items."""
        CANONICALIZE_PASS.apply(CTX, self.workload_constant_100)

    def setup_constant_folding_1000(self) -> None:
        """Setup the constant folding 1000 items benchmark."""
        self.workload_constant_1000 = ConstantFolding.WORKLOAD_CONSTANT_1000.clone()

    def time_constant_folding_1000(self) -> None:
        """Time canonicalizing constant folding for 1000 items."""
        CANONICALIZE_PASS.apply(CTX, self.workload_constant_1000)


class RewritingMicrobenchmarks:
    """Microbenchmarks for rewriting of constant folding."""

    WORKLOAD_CONSTANT_20 = parse_module(CTX, WorkloadBuilder.constant_folding(20))
    region: Region
    const_0: ConstantOp
    const_1: ConstantOp
    add_op: AddiOp
    sub_op: SubiOp
    integer_type: IntegerType

    def setup(self) -> None:
        """Setup the benchmarks."""
        self.region = ConstantFolding.WORKLOAD_CONSTANT_20.clone().body
        self.first_block = self.region.block
        ops = self.region.walk()
        self.const_0 = cast(ConstantOp, next(ops))
        self.const_1 = cast(ConstantOp, next(ops))
        self.add_op = cast(AddiOp, next(ops))
        self.add_op_def = AddiOp.get_irdl_definition()
        self.add_op_construct = VarIRConstruct.OPERAND
        self.add_op_result = self.add_op.result
        self.add_op_result_use = list(self.add_op_result.uses)[0]
        self.sub_op = SubiOp(self.const_1, self.const_0)
        self.sub_op_result = self.sub_op.result
        self.insert_point = InsertPoint.before(self.add_op)
        self.worklist = Worklist()
        self.worklist.push(self.sub_op)
        self.integer_type = IntegerType(64)
        self.integer_attr = IntegerAttr(0, 64)
        self.live_set = LiveSet()
        self.pattern_rewriter = PatternRewriter(self.add_op)


class PatternRewriting(RewritingMicrobenchmarks):
    """Microbenchmarks for general pattern rewriting of constant folding."""

    def time_region_walk(self) -> None:
        """Time `Region.walk`.

        Exercise walking over an IR region, including iterator overhead. This
        is used to populate the worklist for pattern rewriting.
        """
        for block in self.region.walk():
            assert block

    def time_worklist_push(self) -> None:
        """Time `Worklist.push`.

        Exercise adding an operation to the worklist, including lookups in the
        worklist map, relying on operation's `__hash__` function. This is used
        to populate the worklist for pattern rewriting.
        """
        self.worklist.push(self.add_op)

    def time_worklist_pop(self) -> None:
        """Time `Worklist.pop`.

        Exercise removing an operation from the worklist, from the worklist map,
        relying on operation's `__hash__` function. This is used to process
        items sequentially from the worklist as a stack.
        """
        self.worklist.pop()

    def time_get_trait(self) -> None:
        """Time `Operation.get_trait`.

        Exercise getting an operation's trait, including `isinstance` checks and
        the traits iterator. This is used frequently in pattern rewriting, both
        for matching and for the rewriting logic.
        """
        self.add_op.get_trait(HasCanonicalizationPatternsTrait)

    def time_insert_point_before(self) -> None:
        """Time `InsertPoint.before`.

        Measure time to create an `InsertPoint` from the block's parent block.
        """
        InsertPoint.before(self.add_op)

    def ignore_time_pattern_rewriter_insert_op(self) -> None:
        """Time `PatternRewriter.insert_op`.

        Exercise inserting an operation and running any required callbacks. This
        is used to effect the results of rewriting.
        """
        self.pattern_rewriter.insert_op((self.sub_op,), self.insert_point)

    def time_get_variadic_sizes(self) -> None:
        """Time `get_variadic_sizes`.

        Exercise getting the variadic size of an operation, including a
        significant amount of logic to check lengths and types. This is invoked
        whenever an operation argument is defined with `irdl_op_arg_definition`.
        """
        get_variadic_sizes(self.add_op, self.add_op_def, self.add_op_construct)


class Canonicalization(RewritingMicrobenchmarks):
    """Microbenchmarks for canonicalization rewriting of constant folding."""

    def ignore_time_operation_drop_all_references(self) -> None:
        """Time `Operation.drop_all_references`.

        Exercise dropping references to an operation, including removing
        references to its IR uses. This is used when removing unused operations.
        """
        self.add_op.drop_all_references()

    def time_ssavalue_replace_by(self) -> None:
        """Time `SSAValue.replace_by`.

        Exercise replacing an `SSAValue` and all its uses with another
        SSAValue. This is used when removing unused operations.
        """
        self.add_op_result.replace_by(self.sub_op_result)

    def ignore_time_irwithuses_remove_use(self) -> None:
        """Time `IRWithUses.remove_use`.

        Exercise removing references to IR uses. This is used when removing
        unused operations.
        """
        self.add_op_result.remove_use(self.add_op_result_use)

    def time_irwithuses_add_use(self) -> None:
        """Time `IRWithUses.add_use`.

        Exercise adding references to IR uses.
        """
        self.sub_op_result.add_use(self.add_op_result_use)

    def time_ssavalue_name_hint(self) -> None:
        """Time `SSAValue.namehint`.

        Exercise changing an `SSAValue` name hint, which performs regex
        validation at runtime. This is used when removing unused operations.
        """
        self.add_op_result.name_hint = "valid_name"

    def time_handle_operation_removal(self) -> None:
        """Time `PatternRewriter.handle_operation_removal`.

        Exercise removing an operation recursively from regions, and from the
        pattern rewriter worklist. This is used when removing unused operations.
        """
        self.pattern_rewriter.handle_operation_removal(self.add_op)

    def ignore_time_block_detach_op(self) -> None:
        """Time `Block.detach_op`.

        Exercise detaching an operation from a block, including fixing the
        block's doubly linked list. This is used when removing unused
        operations.
        """
        self.region.block.detach_op(self.add_op)

    def time_ssavalue_erase(self) -> None:
        """Time `SSAValue.erase`.

        Exercise erasing an 'SSAValue'. This is used when removing unused
        operations.
        """
        self.add_op_result.erase(safe_erase=False)

    def time_integer_attr_creation(self) -> None:
        """Time `IntegerAttr.__init__`.

        Exercise instantiating an integer attribute, including getting a
        normalised value and constructing the class. This is used when creating
        a new constant operation for folding.
        """
        IntegerAttr(0, 64)

    def time_integer_type_normalized_value(self) -> None:
        """Time `IntegerType.normalized_value`.

        Exercise getting the normalised value for an integer. This is used when creating
        a new constant operation for folding.
        """
        self.integer_type.normalized_value(0)

    def time_integer_attr_verify(self) -> None:
        """Time `IntegerAttr._verify`.

        Exercise verifying an integer attribute. This is used when creating a
        new constant operation for folding.
        """
        self.integer_attr._verify()  # pyright: ignore[reportPrivateUsage]

    def time_operation_create(self) -> None:
        """Time `AddiOp.__init__`.

        Exercising instantiating an add operation of two constant values.
        """
        AddiOp(self.const_1, self.const_0)

    def time_const_evaluate_operand(self) -> None:
        """Time `const_evaluate_operand`.

        Exercise getting the value from a constant.
        """
        const_evaluate_operand(self.add_op.lhs)


class RemoveUnused(RewritingMicrobenchmarks):
    """Microbenchmarks for unused code removal rewriting of constant folding."""

    def time_is_trivially_dead(self) -> None:
        """Time `is_trivially_dead`.

        Exercise checking if an operation has no observable effect. This is
        mostly checking if it would be trivially dead, and alse checking its
        results have no uses.
        """
        is_trivially_dead(self.add_op)

    def time_would_be_trivially_dead(self) -> None:
        """Time `would_be_trivially_dead`.

        Exercise checking if an operation would be dead if all its results were
        dead. This is around half trait checks and half getting result only
        effects.
        """
        would_be_trivially_dead(self.add_op)

    def time_result_only_effects(self) -> None:
        """Time `result_only_effects`.

        Exercise checking if an operation has no observable effect beyond its
        return value. This is mostly getting its effects.
        """
        result_only_effects(self.add_op)

    def time_operation_get_traits_of_type(self) -> None:
        """Time `Operation.get_traits_of_type`.

        Exercise getting all the traits from an operatino of a give type. This
        is mostly `isinstance` checks.
        """
        self.add_op.get_traits_of_type(MemoryEffect)


class RegionDCE(RewritingMicrobenchmarks):
    """Microbenchmarks for region dead-code elimination of constant folding."""

    def time_post_order_iterator(self) -> None:
        """Time `PostOrderIterator`.

        Exercise constructing and using a post-order iterator over a block. This
        is used for dead code elimination.
        """
        post_order_iterator = PostOrderIterator(self.first_block)
        for block in post_order_iterator:
            assert block

    def time_liveset_set_live(self) -> None:
        """Time `LiveSet.set_live`.

        Exercise adding an item to the live set. This is used for dead code
        elimination.
        """
        self.live_set.set_live(self.add_op)

    def time_liveset_delete_dead(self) -> None:
        """Time `LiveSet.delete_dead`.

        Exercise deleting operations not in the live set for a region. This is
        used for dead code elimination.
        """
        self.live_set.delete_dead(self.region, None)


if __name__ == "__main__":
    from bench_utils import Benchmark, profile

    GENERAL = PatternRewriting()
    CONSTANT_FOLDING = ConstantFolding()
    CANONICALIZATION = Canonicalization()
    REMOVE_UNUSED = RemoveUnused()
    REGION_DCE = RegionDCE()
    profile(
        {
            "ConstantFolding.20": Benchmark(
                CONSTANT_FOLDING.time_constant_folding_20,
                CONSTANT_FOLDING.setup_constant_folding_20,
            ),
            "ConstantFoldingSimple.20": Benchmark(
                CONSTANT_FOLDING.time_constant_folding_simple_20,
                CONSTANT_FOLDING.setup_constant_folding_20,
            ),
            "ConstantFoldingSpecialised.20": Benchmark(
                CONSTANT_FOLDING.time_constant_folding_specialised_20,
                CONSTANT_FOLDING.setup_constant_folding_20,
            ),
            "ConstantFolding.100": Benchmark(
                CONSTANT_FOLDING.time_constant_folding_100,
                CONSTANT_FOLDING.setup_constant_folding_100,
            ),
            "ConstantFolding.1000": Benchmark(
                CONSTANT_FOLDING.time_constant_folding_1000,
                CONSTANT_FOLDING.setup_constant_folding_1000,
            ),
            # ================================================================ #
            "General.region_walk": Benchmark(GENERAL.time_region_walk, GENERAL.setup),
            "General.worklist_push": Benchmark(
                GENERAL.time_worklist_push,
                GENERAL.setup,
            ),
            "General.worklist_pop": Benchmark(
                GENERAL.time_worklist_pop,
                GENERAL.setup,
            ),
            "General.get_trait": Benchmark(
                GENERAL.time_get_trait,
                GENERAL.setup,
            ),
            "General.insert_point_before": Benchmark(
                GENERAL.time_insert_point_before,
                GENERAL.setup,
            ),
            "General.pattern_rewriter_insert_op": Benchmark(
                GENERAL.ignore_time_pattern_rewriter_insert_op,
                GENERAL.setup,
            ),
            "General.get_variadic_sizes": Benchmark(
                GENERAL.time_get_variadic_sizes, GENERAL.setup
            ),
            "Canonicalization.operation_drop_all_references": Benchmark(
                CANONICALIZATION.ignore_time_operation_drop_all_references,
                CANONICALIZATION.setup,
            ),
            "Canonicalization.ssavalue_replace_by": Benchmark(
                CANONICALIZATION.time_ssavalue_replace_by,
                CANONICALIZATION.setup,
            ),
            "Canonicalization.irwithuses_remove_use": Benchmark(
                CANONICALIZATION.ignore_time_irwithuses_remove_use,
                CANONICALIZATION.setup,
            ),
            "Canonicalization.irwithuses_add_use": Benchmark(
                CANONICALIZATION.time_irwithuses_add_use,
                CANONICALIZATION.setup,
            ),
            "Canonicalization.ssavalue_name_hint": Benchmark(
                CANONICALIZATION.time_ssavalue_name_hint,
                CANONICALIZATION.setup,
            ),
            "Canonicalization.handle_operation_removal": Benchmark(
                CANONICALIZATION.time_handle_operation_removal,
                CANONICALIZATION.setup,
            ),
            "Canonicalization.block_detach_op": Benchmark(
                CANONICALIZATION.ignore_time_block_detach_op,
                CANONICALIZATION.setup,
            ),
            "Canonicalization.ssavalue_erase": Benchmark(
                CANONICALIZATION.time_ssavalue_erase,
                CANONICALIZATION.setup,
            ),
            "Canonicalization.integer_attr_creation": Benchmark(
                CANONICALIZATION.time_integer_attr_creation,
                CANONICALIZATION.setup,
            ),
            "Canonicalization.integer_type_normalized_value": Benchmark(
                CANONICALIZATION.time_integer_type_normalized_value,
                CANONICALIZATION.setup,
            ),
            "Canonicalization.integer_attr_verify": Benchmark(
                CANONICALIZATION.time_integer_attr_verify,
                CANONICALIZATION.setup,
            ),
            "Canonicalization.operation_create": Benchmark(
                CANONICALIZATION.time_operation_create,
                CANONICALIZATION.setup,
            ),
            "Canonicalization.const_evaluate_operand": Benchmark(
                CANONICALIZATION.time_const_evaluate_operand,
                CANONICALIZATION.setup,
            ),
            "RemoveUnused.is_trivially_dead": Benchmark(
                REMOVE_UNUSED.time_is_trivially_dead,
                REMOVE_UNUSED.setup,
            ),
            "RemoveUnused.would_be_trivially_dead": Benchmark(
                REMOVE_UNUSED.time_would_be_trivially_dead,
                REMOVE_UNUSED.setup,
            ),
            "RemoveUnused.result_only_effects": Benchmark(
                REMOVE_UNUSED.time_result_only_effects,
                REMOVE_UNUSED.setup,
            ),
            "RemoveUnused.operation_get_traits_of_type": Benchmark(
                REMOVE_UNUSED.time_operation_get_traits_of_type,
                REMOVE_UNUSED.setup,
            ),
            "RegionDCE.post_order_iterator": Benchmark(
                REGION_DCE.time_post_order_iterator,
                REGION_DCE.setup,
            ),
            "RegionDCE.liveset_set_live": Benchmark(
                REGION_DCE.time_liveset_set_live,
                REGION_DCE.setup,
            ),
            "RegionDCE.liveset_delete_dead": Benchmark(
                REGION_DCE.time_liveset_delete_dead,
                REGION_DCE.setup,
            ),
        }
    )



from collections.abc import Mapping
from typing import TypeAlias

import pytest

from xdsl.dialects.builtin import (
    ArrayAttr,
    Float32Type,
    FloatAttr,
    IntAttr,
    StringAttr,
    i32,
)
from xdsl.dialects.dlti import (
    DataLayoutEntryAttr,
    DataLayoutSpecAttr,
    DLTIEntryMap,
    MapAttr,
    TargetDeviceSpecAttr,
    TargetSystemSpecAttr,
)
from xdsl.ir import Attribute, TypeAttribute
from xdsl.utils.exceptions import VerifyException

DictValueType: TypeAlias = Mapping[
    StringAttr | TypeAttribute | str, "Attribute | str | int | DictValueType"
]


def test_data_layout_entry():
    # Test passing strings for key and value
    entry = DataLayoutEntryAttr("test", "V")
    assert entry.key == StringAttr("test")
    assert entry.value == StringAttr("V")

    # Test passing string for key and int for value
    entry = DataLayoutEntryAttr("test2", 12)
    assert entry.key == StringAttr("test2")
    assert entry.value == IntAttr(12)

    # Test passing string for key and float for value
    entry = DataLayoutEntryAttr("test3", FloatAttr(99.45, Float32Type()))
    assert entry.key == StringAttr("test3")
    assert entry.value == FloatAttr(99.45, Float32Type())

    # Test passing type for key and string for value
    entry = DataLayoutEntryAttr(i32, "test")
    assert entry.key == i32
    assert entry.value == StringAttr("test")

    # Test passing StringAttr for key and value
    entry = DataLayoutEntryAttr(StringAttr("k"), StringAttr("v"))
    assert entry.key == StringAttr("k")
    assert entry.value == StringAttr("v")


def test_incorrect_data_layout_entry():
    with pytest.raises(VerifyException):
        DataLayoutEntryAttr("", "V")


@pytest.mark.parametrize(
    "cls", [DataLayoutSpecAttr, TargetDeviceSpecAttr, TargetSystemSpecAttr, MapAttr]
)
@pytest.mark.parametrize(
    "entries",
    [
        [StringAttr("value1"), IntAttr(23), IntAttr(43)],
        [IntAttr(23), FloatAttr(2.4, Float32Type())],
    ],
)
def test_entry_maps(
    cls: type[DLTIEntryMap],
    entries: list[Attribute],
):
    # Comparison
    comparison_attr = cls.new(
        (
            ArrayAttr(
                [
                    DataLayoutEntryAttr("key_" + str(idx), e)
                    for idx, e in enumerate(entries)
                ]
            ),
        )
    )

    # Test initialisation from a dictionary
    attr_one = cls({"key_" + str(idx): e for idx, e in enumerate(entries)})
    assert comparison_attr == attr_one
    # Test initialisation from an array attribute of data layout entries
    attr_two = cls(
        ArrayAttr(
            [DataLayoutEntryAttr("key_" + str(idx), e) for idx, e in enumerate(entries)]
        )
    )
    assert comparison_attr == attr_two


def test_map_attr_embedded_dict():
    first_map = MapAttr({"key": {"key": {"key": "value", i32: 21}}})
    second_map = first_map["key"]
    assert isinstance(second_map, MapAttr)
    third_map = second_map["key"]
    assert isinstance(third_map, MapAttr)
    assert third_map["key"] == StringAttr("value")
    assert third_map[i32] == IntAttr(21)


@pytest.mark.parametrize(
    "cls", [DataLayoutSpecAttr, TargetDeviceSpecAttr, TargetSystemSpecAttr, MapAttr]
)
def test_duplicate_data_layout_map_entries(
    cls: type[DLTIEntryMap],
):
    with pytest.raises(VerifyException):
        cls(ArrayAttr([DataLayoutEntryAttr("k", "v"), DataLayoutEntryAttr("k", 12)]))


@pytest.mark.parametrize(
    "cls", [DataLayoutSpecAttr, TargetDeviceSpecAttr, TargetSystemSpecAttr, MapAttr]
)
def test_not_found_map_entry_lookup(
    cls: type[DLTIEntryMap],
):
    attr = cls(
        ArrayAttr([DataLayoutEntryAttr("key1", "v"), DataLayoutEntryAttr("key2", 12)])
    )
    with pytest.raises(KeyError):
        attr["key3"]



import pytest

from xdsl.context import Context
from xdsl.dialects.arith import Arith
from xdsl.dialects.builtin import Builtin
from xdsl.dialects.cf import Cf
from xdsl.dialects.func import Func
from xdsl.dialects.test import Test
from xdsl.ir import Operation
from xdsl.parser import Parser
from xdsl.rewriting.composable_rewriting.immutable_ir.immutable_ir import (  # noqa
    get_immutable_copy,
)

program_region = """
"builtin.module"() ({
  %0 = "arith.constant"() <{"value" = 1 : i32}> : () -> i32
}) : () -> ()
"""

program_region_2 = """
"builtin.module"() ({
  %0 = "arith.constant"() <{"value" = 2 : i32}> : () -> i32
}) : () -> ()
"""

program_region_2_diff_name = """
"builtin.module"() ({
  %0 = "arith.constant"() <{"value" = 2 : i32}> : () -> i32
}) : () -> ()
"""

program_region_2_diff_type = """
"builtin.module"() ({
  %0 = "arith.constant"() <{"value" = 2 : i64}> : () -> i64
}) : () -> ()
"""

program_add = """
"builtin.module"() ({
  %0 = "arith.constant"() <{"value" = 1 : i32}> : () -> i32
  %1 = "arith.constant"() <{"value" = 2 : i32}> : () -> i32
  %2 = "arith.addi"(%0, %1) : (i32, i32) -> i32
}) : () -> ()
"""

program_add_2 = """
"builtin.module"() ({
  %0 = "arith.constant"() <{"value" = 1 : i32}> : () -> i32
  %1 = "arith.constant"() <{"value" = 2 : i32}> : () -> i32
  %2 = "arith.addi"(%1, %0) : (i32, i32) -> i32
}) : () -> ()
"""

program_func = """
"builtin.module"() ({
  "func.func"() <{"sym_name" = "test", "function_type" = (i32, i32) -> i32, "sym_visibility" = "private"}> ({
  ^0(%0 : i32, %1 : i32):
    %2 = "arith.addi"(%0, %1) : (i32, i32) -> i32
    "func.return"(%2) : (i32) -> ()
  }) : () -> ()
}) : () -> ()
"""

program_successors = """
"builtin.module"() ({
  "func.func"() <{"sym_name" = "unconditional_br", "function_type" = () -> (), "sym_visibility" = "private"}> ({
  ^0:
    "cf.br"() [^1] : () -> ()
  ^1:
    "cf.br"() [^0] : () -> ()
  }) : () -> ()
}) : () -> ()
"""

program_attr_and_prop = """
"builtin.module"() ({
  "test.op"() <{"prop1" = i32}> {"attr1" = i64} : () -> ()
}) : () -> ()
"""


@pytest.mark.parametrize(
    "program_str",
    [
        program_region,
        program_region_2,
        program_region_2_diff_type,
        program_region_2_diff_name,
        program_add,
        program_add_2,
        program_func,
        program_successors,
        program_attr_and_prop,
    ],
)
def test_immutable_ir(program_str: str):
    ctx = Context()
    ctx.load_dialect(Test)
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Func)
    ctx.load_dialect(Arith)
    ctx.load_dialect(Cf)

    parser = Parser(ctx, program_str)
    program: Operation = parser.parse_op()
    immutable_program = get_immutable_copy(program)
    mutable_program = immutable_program.to_mutable()

    assert program.is_structurally_equivalent(mutable_program)



class LazyVersion:
    """
    Resolving the version dynamically is slow, hence this lazy wrapper to get the version
    only when needed.
    """

    _version: str | None

    def __init__(self):
        self._version = None

    def __str__(self):
        if self._version is None:
            from importlib.metadata import PackageNotFoundError, version

            try:
                self._version = version("xdsl")
            except PackageNotFoundError:
                # package is not installed
                self._version = "Error: xDSL not installed"

        return self._version


__version__ = LazyVersion()



from xdsl.dialects import test
from xdsl.pattern_rewriter import Worklist


def test_worklist_push_pop():
    """Test push/pop operations on the worklist."""
    op1 = test.TestOp()
    op2 = test.TestOp()
    op3 = test.TestOp()
    op4 = test.TestOp()

    worklist = Worklist()

    worklist.push(op1)
    worklist.push(op3)
    worklist.push(op2)
    worklist.push(op4)

    assert worklist.pop() is op4
    assert worklist.pop() is op2
    worklist.push(op4)
    assert worklist.pop() is op4
    assert worklist.pop() is op3
    assert worklist.pop() is op1
    assert worklist.is_empty()
    assert worklist.pop() is None
    assert worklist.is_empty()


def test_worklist_push_already_inserted():
    """Test push of an already inserted operation."""

    op1 = test.TestOp()
    op2 = test.TestOp()

    worklist = Worklist()
    worklist.push(op1)
    worklist.push(op2)

    worklist.push(op1)

    assert worklist.pop() is op2
    assert worklist.pop() is op1
    assert worklist.is_empty()
    assert worklist.pop() is None
    assert worklist.is_empty()


def test_worklist_remove():
    """Test remove operation."""
    op1 = test.TestOp()
    op2 = test.TestOp()
    op3 = test.TestOp()
    op4 = test.TestOp()

    worklist = Worklist()

    worklist.push(op1)
    worklist.push(op3)
    worklist.push(op2)
    worklist.push(op4)

    worklist.remove(op1)
    worklist.remove(op2)
    worklist.remove(op2)

    assert worklist.pop() is op4
    assert worklist.pop() is op3
    assert worklist.is_empty()
    assert worklist.pop() is None
    assert worklist.is_empty()



"""
A pass that applies the interpreter to operations with no side effects where all the
inputs are constant, replacing the computation with a constant value.
"""

from dataclasses import dataclass

from xdsl.context import Context
from xdsl.dialects import builtin, func, pdl, test
from xdsl.ir import Block, Region
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)


@dataclass
class FuncOpToPdlRewritePattern(RewritePattern):
    """Rewrite pattern that transforms a function into a PDL rewrite operation."""

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: func.FuncOp, rewriter: PatternRewriter):
        pdl_root = test.TestPureOp(result_types=(pdl.OperationType(),))
        op.detach_region(func_body := op.regions[0])
        rewrite_root = pdl_root.results[0]
        for arg in func_body.block.args:
            arg.replace_by(rewrite_root)
            func_body.block.erase_arg(arg)

        pdl_pattern = pdl.PatternOp(
            benefit=1,
            sym_name=op.sym_name.data,
            body=Region(
                Block([pdl_root, pdl.RewriteOp(root=rewrite_root, body=func_body)])
            ),
        )
        rewriter.replace_matched_op(pdl_pattern)


@dataclass
class ReturnOpToPdlRewritePattern(RewritePattern):
    """Rewrite pattern that erases return ops in PDL rewrite operations."""

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: func.ReturnOp, rewriter: PatternRewriter):
        rewriter.erase_matched_op()


class FuncToPdlRewrite(ModulePass):
    """
    A pass that transforms a function into a PDL rewrite operation.
    """

    name = "func-to-pdl-rewrite"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        """Apply the pass."""
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    FuncOpToPdlRewritePattern(),
                    ReturnOpToPdlRewritePattern(),
                ]
            )
        ).rewrite_module(op)



import pytest

from xdsl.dialects.builtin import (
    IntegerAttr,
    IntegerType,
    i1,
    i32,
)
from xdsl.dialects.seq import (
    ClockDividerOp,
    CompRegOp,
    clock,
)
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.test_value import create_ssa_value


def test_clockdivider_verify():
    clock_div = ClockDividerOp(
        create_ssa_value(clock),
        IntegerAttr(512, i32),
    )
    with pytest.raises(
        VerifyException,
        match="Operation does not verify: pow2 has to be an 8-bit signless integer",
    ):
        clock_div.verify()


def test_compreg_builder():
    data_val = create_ssa_value(IntegerType(5))
    bool_val = create_ssa_value(i1)
    clock_val = create_ssa_value(clock)

    CompRegOp(data_val, clock_val).verify()

    with_reset = CompRegOp(data_val, clock_val, (bool_val, data_val))
    with_reset.verify()
    assert with_reset.reset is not None



from xdsl.builder import ImplicitBuilder
from xdsl.context import Context
from xdsl.dialects import test
from xdsl.dialects.builtin import IntAttr
from xdsl.ir import Block, Region


def test_preorder_walk():
    ctx = Context()
    ctx.load_dialect(test.Test)

    # Create a region with multiple blocks
    outer_region = Region([Block(), Block()])
    outer_block1, outer_block2 = outer_region.blocks

    # Create nested regions with multiple blocks
    nested_region1 = Region([Block(), Block()])
    nested_block1, nested_block2 = nested_region1.blocks

    nested_region2 = Region([Block(), Block()])
    nested_block3, nested_block4 = nested_region2.blocks

    # Create a third region for multi-region op
    nested_region3 = Region([Block(), Block()])
    nested_block5, nested_block6 = nested_region3.blocks

    # Build operations with nested regions
    root_op = test.TestOp(regions=(outer_region,))

    with ImplicitBuilder(outer_block1):
        test.TestOp(attributes={"id": IntAttr(1)})
        with ImplicitBuilder(nested_block1):
            test.TestOp(attributes={"id": IntAttr(3)})
            test.TestTermOp()
        with ImplicitBuilder(nested_block2):
            test.TestOp(attributes={"id": IntAttr(4)})
            test.TestTermOp()
        test.TestOp(regions=(nested_region1,))
        test.TestTermOp()

    with ImplicitBuilder(outer_block2):
        test.TestOp(attributes={"id": IntAttr(2)})
        with ImplicitBuilder(nested_block3):
            test.TestOp(attributes={"id": IntAttr(5)})
            test.TestTermOp()
        with ImplicitBuilder(nested_block4):
            test.TestOp(attributes={"id": IntAttr(6)})
            test.TestTermOp()
        # Add multi-region op
        with ImplicitBuilder(nested_block5):
            test.TestOp(attributes={"id": IntAttr(7)})
            test.TestTermOp()
        with ImplicitBuilder(nested_block6):
            test.TestOp(attributes={"id": IntAttr(8)})
            test.TestTermOp()
        test.TestOp(regions=(nested_region2, nested_region3))
        test.TestTermOp()

    assert tuple(root_op.walk_blocks()) == (
        outer_block1,
        nested_block1,
        nested_block2,
        outer_block2,
        nested_block3,
        nested_block4,
        nested_block5,
        nested_block6,
    )

    assert tuple(outer_block1.walk_blocks()) == (
        outer_block1,
        nested_block1,
        nested_block2,
    )

    assert tuple(outer_block2.walk_blocks()) == (
        outer_block2,
        nested_block3,
        nested_block4,
        nested_block5,
        nested_block6,
    )

    assert tuple(root_op.walk_blocks(reverse=True)) == (
        nested_block6,
        nested_block5,
        nested_block4,
        nested_block3,
        outer_block2,
        nested_block2,
        nested_block1,
        outer_block1,
    )

    assert tuple(outer_block1.walk_blocks(reverse=True)) == (
        nested_block2,
        nested_block1,
        outer_block1,
    )

    assert tuple(outer_block2.walk_blocks(reverse=True)) == (
        nested_block6,
        nested_block5,
        nested_block4,
        nested_block3,
        outer_block2,
    )



from xdsl.context import Context
from xdsl.dialects import builtin
from xdsl.ir import Operation
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
)
from xdsl.traits import HasCanonicalizationPatternsTrait
from xdsl.transforms.dead_code_elimination import RemoveUnusedOperations, region_dce


class CanonicalizationRewritePattern(RewritePattern):
    """Rewrite pattern that applies a canonicalization pattern."""

    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter, /):
        trait = op.get_trait(HasCanonicalizationPatternsTrait)
        if trait is None:
            return
        patterns = trait.get_canonicalization_patterns()
        if len(patterns) == 1:
            patterns[0].match_and_rewrite(op, rewriter)
            return
        GreedyRewritePatternApplier(list(patterns)).match_and_rewrite(op, rewriter)


class CanonicalizePass(ModulePass):
    """
    Applies all canonicalization patterns.
    """

    name = "canonicalize"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        pattern = GreedyRewritePatternApplier(
            [RemoveUnusedOperations(), CanonicalizationRewritePattern()]
        )
        PatternRewriteWalker(pattern, post_walk_func=region_dce).rewrite_module(op)



from dataclasses import dataclass, field

from ordered_set import OrderedSet

from xdsl.builder import Builder, InsertPoint
from xdsl.context import Context
from xdsl.dialects import builtin, riscv, riscv_func
from xdsl.dialects.riscv import (
    FloatRegisterType,
    IntRegisterType,
    Registers,
    RISCVRegisterType,
)
from xdsl.passes import ModulePass


@dataclass(frozen=True)
class PrologueEpilogueInsertion(ModulePass):
    """
    Pass inserting a prologue and epilogue according to the RISC-V ABI.
    The prologues and epilogues are responsible for saving any callee-preserved
    registers.
    In RISC-V these are 's0' to 's11' and 'fs0' to `fs11'.
    The stack pointer 'sp' must also be restored to its original value.

    This pass should be run late in the pipeline after register allocation.
    It does not itself require register allocation nor invalidate the result of the
    register allocator.
    """

    name = "riscv-prologue-epilogue-insertion"
    xlen: int = field(default=4)
    flen: int = field(default=8)

    def _process_function(self, func: riscv_func.FuncOp) -> None:
        # Find all callee-preserved registers that are clobbered. We define clobbered
        # as it being the result of some operation and therefore written to.
        used_callee_preserved_registers = OrderedSet(
            res.type
            for op in func.walk()
            if not isinstance(op, riscv.GetRegisterOp | riscv.GetFloatRegisterOp)
            for res in op.results
            if isinstance(res.type, IntRegisterType | FloatRegisterType)
            if res.type in Registers.S or res.type in Registers.FS
        )

        if not used_callee_preserved_registers:
            return

        def get_register_size(r: RISCVRegisterType):
            if isinstance(r, IntRegisterType):
                return self.xlen
            return self.flen

        # Build the prologue at the beginning of the function.
        builder = Builder(InsertPoint.at_start(func.body.blocks[0]))
        sp_register = builder.insert(riscv.GetRegisterOp(Registers.SP))
        stack_size = sum(get_register_size(r) for r in used_callee_preserved_registers)
        builder.insert(riscv.AddiOp(sp_register, -stack_size, rd=Registers.SP))
        offset = 0
        for reg in used_callee_preserved_registers:
            if isinstance(reg, IntRegisterType):
                reg_op = builder.insert(riscv.GetRegisterOp(reg))
                op = riscv.SwOp(rs1=sp_register, rs2=reg_op, immediate=offset)
            else:
                reg_op = builder.insert(riscv.GetFloatRegisterOp(reg))
                op = riscv.FSdOp(rs1=sp_register, rs2=reg_op, immediate=offset)

            builder.insert(op)
            offset += get_register_size(reg)

        # Now build the epilogue right before every return operation.
        for block in func.body.blocks:
            ret_op = block.last_op
            if not isinstance(ret_op, riscv_func.ReturnOp):
                continue

            builder = Builder(InsertPoint.before(ret_op))
            offset = 0
            for reg in used_callee_preserved_registers:
                if isinstance(reg, IntRegisterType):
                    op = riscv.LwOp(rs1=sp_register, rd=reg, immediate=offset)
                else:
                    op = riscv.FLdOp(rs1=sp_register, rd=reg, immediate=offset)
                builder.insert(op)
                offset += get_register_size(reg)

            builder.insert(riscv.AddiOp(sp_register, stack_size, rd=Registers.SP))

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        for func in op.walk():
            if not isinstance(func, riscv_func.FuncOp):
                continue

            if len(func.body.blocks) == 0:
                continue

            self._process_function(func)



from typing import cast

from xdsl.backend.register_type import RegisterType
from xdsl.backend.utils import cast_to_regs
from xdsl.dialects import x86
from xdsl.dialects.builtin import (
    FixedBitwidthType,
    IndexType,
    ShapedType,
    VectorType,
)
from xdsl.dialects.x86.register import X86VectorRegisterType
from xdsl.ir import Attribute, SSAValue
from xdsl.pattern_rewriter import PatternRewriter
from xdsl.utils.exceptions import DiagnosticException


def vector_type_to_register_type(
    value_type: VectorType,
    arch: str,
) -> X86VectorRegisterType:
    vector_num_elements = value_type.element_count()
    element_type = cast(FixedBitwidthType, value_type.get_element_type())
    element_size = element_type.bitwidth
    vector_size = vector_num_elements * element_size
    # Choose the x86 vector register according to the
    # target architecture and the abstract vector size
    if vector_size == 128:
        vect_reg_type = x86.register.UNALLOCATED_SSE
    elif vector_size == 256 and (arch == "avx2" or arch == "avx512"):
        vect_reg_type = x86.register.UNALLOCATED_AVX2
    elif vector_size == 512 and arch == "avx512":
        vect_reg_type = x86.register.UNALLOCATED_AVX512
    else:
        raise DiagnosticException(
            "The vector size and target architecture are inconsistent."
        )
    return vect_reg_type


def scalar_type_to_register_type(value_type: Attribute) -> type[RegisterType]:
    assert not isinstance(value_type, ShapedType)
    if (
        isinstance(value_type, FixedBitwidthType) and value_type.bitwidth <= 64
    ) or isinstance(value_type, IndexType):
        return x86.register.GeneralRegisterType
    else:
        raise DiagnosticException("Not implemented for bitwidth larger than 64.")


def cast_operands_to_regs(rewriter: PatternRewriter) -> list[SSAValue[Attribute]]:
    new_ops, new_operands = cast_to_regs(
        values=rewriter.current_operation.operands,
        register_map=scalar_type_to_register_type,
    )
    rewriter.insert_op_before_matched_op(new_ops)
    return new_operands



import ast
import functools
from collections.abc import Callable
from contextlib import AbstractContextManager
from dataclasses import dataclass, field
from inspect import currentframe, getsource
from sys import _getframe  # pyright: ignore[reportPrivateUsage]
from types import FrameType
from typing import Any, NamedTuple

from xdsl.context import Context
from xdsl.dialects.builtin import ModuleOp
from xdsl.frontend.pyast.program import FrontendProgram, P, PyASTProgram, R
from xdsl.frontend.pyast.utils.builder import PyASTBuilder
from xdsl.frontend.pyast.utils.python_code_check import PythonCodeCheck
from xdsl.frontend.pyast.utils.type_conversion import FunctionRegistry, TypeRegistry
from xdsl.ir import Dialect, Operation, TypeAttribute
from xdsl.passes import ModulePass, PassPipeline
from xdsl.transforms.desymref import FrontendDesymrefyPass


class FuncInfo(NamedTuple):
    """Information about a decorated function being generated into IR."""

    file: str
    """The path of the file containing the function."""

    globals: dict[str, Any]
    """The globals defined in that file up to the point of function definition."""

    ast: ast.FunctionDef
    """The Python AST representation of the function."""


def default_pipeline_callback(
    _previous_pass: ModulePass, module: ModuleOp, _next_pass: ModulePass
) -> None:
    """Default callback to verify the module after each transformation pass."""
    module.verify()


@dataclass
class PyASTContext:
    """Encapsulate the mapping between Python and IR types and operations."""

    type_registry: TypeRegistry = field(default_factory=TypeRegistry)
    """Mappings between source code and IR type."""

    function_registry: FunctionRegistry = field(default_factory=FunctionRegistry)
    """Mappings between functions and their operation types."""

    post_transforms: list[ModulePass] = field(
        default_factory=lambda: [FrontendDesymrefyPass()]
    )
    """An ordered list of passes to apply to the built module."""

    post_callback: Callable[[ModulePass, ModuleOp, ModulePass], None] | None = (
        default_pipeline_callback
    )
    """Callback to run between post transforms."""

    ir_context: Context = field(
        default_factory=lambda: Context(allow_unregistered=True)
    )
    """The xDSL context to use when applying transformations to the built module."""

    def register_type(
        self,
        source_type: type,
        ir_type: TypeAttribute,
    ) -> None:
        """Associate a type in the source code with its type in the IR."""
        self.type_registry.insert(source_type, ir_type)

    def register_function(
        self, function: Callable[..., Any], ir_constructor: Callable[..., Operation]
    ) -> None:
        """Associate a method on an object in the source code with its IR implementation."""
        self.function_registry.insert(function, ir_constructor)

    def register_post_transform(self, transform: ModulePass) -> None:
        """Add a module pass to be run on the generated IR."""
        self.post_transforms.append(transform)

    def register_dialect(self, dialect: Dialect) -> None:
        """Add a dialect to the context used for transformation."""
        self.ir_context.load_dialect(dialect)

    @property
    def pass_pipeline(self) -> PassPipeline:
        """Get a pass pipeline from the context state."""
        return PassPipeline(tuple(self.post_transforms), self.post_callback)

    @classmethod
    def _get_func_info(
        cls,
        current_frame: FrameType | None,
        func: Callable[P, R],
    ) -> FuncInfo:
        """Get information about the decorated function."""
        # Get the correct function frame from the call stack
        assert current_frame is not None
        func_frame = current_frame.f_back
        assert func_frame is not None

        # Get the required information about the function from the frame
        func_file = func_frame.f_code.co_filename
        func_globals = func_frame.f_globals

        # Retrieve the AST for the function body, without the decorator
        func_ast = ast.parse(getsource(func.__code__)).body[0]
        assert isinstance(func_ast, ast.FunctionDef)
        assert func_ast.name == func.__name__
        assert len(func_ast.decorator_list) == 1
        func_ast.decorator_list = []

        # Return the information about the function
        return FuncInfo(func_file, func_globals, func_ast)

    @classmethod
    def _get_wrapped_program(
        cls, func: Callable[P, R], builder: PyASTBuilder
    ) -> PyASTProgram[P, R]:
        """Return a PyAST program for this function with the builder."""
        program = PyASTProgram[P, R](
            name=func.__name__,
            func=func,
            _builder=builder,
        )
        functools.update_wrapper(program, func)
        assert program.__doc__ == func.__doc__
        return program

    def parse_program(self, func: Callable[P, R]) -> PyASTProgram[P, R]:
        """Get a program wrapper by decorating a function."""
        func_file, func_globals, func_ast = self._get_func_info(currentframe(), func)
        builder = PyASTBuilder(
            type_registry=self.type_registry,
            function_registry=self.function_registry,
            file=func_file,
            globals=func_globals,
            function_ast=func_ast,
            build_context=self.ir_context,
            post_transforms=self.pass_pipeline,
        )
        return self._get_wrapped_program(func, builder)


@dataclass
class CodeContext(AbstractContextManager[Any]):
    """
    The CodeContext with block marks the scope in which the code in the custom
    DSL can be written. This code will be translated to xDSL/MLIR.
    """

    program: FrontendProgram
    """
    Underlying frontend program which can be compiled and translated to
    xDSL/MLIR.
    """

    def __enter__(self) -> None:
        # First, get the Python AST from the code.
        frame = _getframe(1)
        self.program.file = frame.f_code.co_filename
        src = getsource(frame)
        python_ast = ast.parse(src)

        # Get all the global information and record it as well. In particular,
        # this contains all the imports.
        self.program.globals = frame.f_globals

        # Find where the program starts.
        for node in ast.walk(python_ast):
            if (
                isinstance(node, ast.With)
                and node.lineno == frame.f_lineno - frame.f_code.co_firstlineno + 1
            ):
                # Found the program AST. Store it for later compilation or
                # execution.
                self.program.stmts = node.body

    def __exit__(self, *args: object):
        # Having proccessed all the code in the context, check it is well-formed
        # and can be compiled/executed. Additionally, record it for subsequent code generation.
        assert self.program.stmts is not None
        self.program.functions_and_blocks = PythonCodeCheck.run(
            self.program.stmts, self.program.file
        )



import abc
from collections.abc import Sequence
from typing import cast

from ordered_set import OrderedSet

from xdsl.backend.register_stack import RegisterStack
from xdsl.backend.register_type import RegisterType
from xdsl.ir import Attribute, Block, SSAValue
from xdsl.rewriter import Rewriter
from xdsl.utils.exceptions import DiagnosticException


class ValueAllocator:
    """
    Base class for register allocators.
    """

    available_registers: RegisterStack
    register_base_class: type[RegisterType]
    new_value_by_old_value: dict[SSAValue, SSAValue]

    def __init__(
        self,
        available_registers: RegisterStack,
        register_base_class: type[RegisterType],
    ) -> None:
        self.available_registers = available_registers
        self.register_base_class = register_base_class
        self.new_value_by_old_value = {}

    def new_type_for_value(self, reg: SSAValue) -> RegisterType | None:
        if isinstance(reg.type, self.register_base_class) and not reg.type.is_allocated:
            return self.available_registers.pop(type(reg.type))

    def _replace_value_with_new_type(
        self, val: SSAValue, new_type: Attribute
    ) -> SSAValue:
        new_val = Rewriter.replace_value_with_new_type(val, new_type)
        self.new_value_by_old_value[val] = new_val
        return new_val

    def allocate_value(self, val: SSAValue) -> SSAValue | None:
        """
        Allocate a register if not already allocated.
        """
        if val in self.new_value_by_old_value:
            return
        new_type = self.new_type_for_value(val)
        if new_type is not None:
            return self._replace_value_with_new_type(val, new_type)

    def allocate_values_same_reg(self, vals: Sequence[SSAValue]) -> bool:
        """
        Allocates the values passed in to the same register.
        If some of the values are already allocated, they must be allocated to the same
        register, and unallocated values are then allocated to this register.
        If the values passed in are already allocated to differing registers, a
        `DiagnosticException` is raised.
        """
        reg_types = set(val.type for val in vals)
        assert all(
            isinstance(reg_type, self.register_base_class) for reg_type in reg_types
        )
        reg_types = cast(set[RegisterType], reg_types)

        match len(reg_types):
            case 0:
                # No inputs, nothing to do
                return False
            case 1:
                # Single input, may already be allocated
                (reg_type,) = reg_types
                if reg_type.is_allocated:
                    return False
                else:
                    reg_type = self.available_registers.pop(type(reg_type))
            case 2:
                # Two inputs, either one is allocated or two
                reg_type_0, reg_type_1 = reg_types
                if reg_type_0.is_allocated:
                    if reg_type_1.is_allocated:
                        reg_names = [f"{reg_type}" for reg_type in reg_types]
                        reg_names.sort()
                        raise DiagnosticException(
                            f"Cannot allocate registers to the same register {reg_names}"
                        )
                    else:
                        reg_type = reg_type_0
                else:
                    reg_type = reg_type_1
            case _:
                # More than one input is allocated, meaning we can't allocate them to be
                # the same, error.
                reg_names = [f"{reg_type}" for reg_type in reg_types]
                reg_names.sort()
                raise DiagnosticException(
                    f"Cannot allocate registers to the same register {reg_names}"
                )

        did_allocate = False

        for val in vals:
            if val.type != reg_type:
                self._replace_value_with_new_type(val, reg_type)
                did_allocate = True

        return did_allocate

    def free_value(self, val: SSAValue) -> None:
        if isinstance(val.type, self.register_base_class) and val.type.is_allocated:
            self.available_registers.push(val.type)


def _live_ins_per_block(
    block: Block, acc: dict[Block, OrderedSet[SSAValue]]
) -> OrderedSet[SSAValue]:
    res = OrderedSet[SSAValue]([])

    for op in reversed(block.ops):
        # Remove values defined in the block
        # We are traversing backwards, so cannot use the value removed here again
        res.difference_update(op.results)
        # Add values used in the block
        res.update(op.operands)

        # Process inner blocks
        for region in op.regions:
            for inner in region.blocks:
                # Add the values used in the inner block
                res.update(_live_ins_per_block(inner, acc))

    # Remove the block arguments
    res.difference_update(block.args)

    acc[block] = res

    return res


def live_ins_per_block(block: Block) -> dict[Block, OrderedSet[SSAValue]]:
    """
    Returns a mapping from a block to the set of values used in it but defined outside of
    it.
    """
    res: dict[Block, OrderedSet[SSAValue]] = {}
    _ = _live_ins_per_block(block, res)
    return res


class BlockAllocator(ValueAllocator, abc.ABC):
    """
    Abstract base class for allocators that can process blocks at a time.
    """

    live_ins_per_block: dict[Block, OrderedSet[SSAValue]]

    def __init__(
        self,
        available_registers: RegisterStack,
        register_base_class: type[RegisterType],
    ) -> None:
        self.live_ins_per_block = {}
        super().__init__(available_registers, register_base_class)

    @abc.abstractmethod
    def allocate_block(self, block: Block):
        """
        For each operation in the block, allocate the registers.
        """



import builtins
import re
from io import StringIO
from typing import cast

import pytest

from xdsl.context import Context
from xdsl.dialect_interfaces import OpAsmDialectInterface
from xdsl.dialects.builtin import (
    ArrayAttr,
    Builtin,
    DictionaryAttr,
    FileLineColLoc,
    FloatAttr,
    IntAttr,
    IntegerAttr,
    IntegerType,
    StringAttr,
    SymbolRefAttr,
    UnknownLoc,
    i32,
)
from xdsl.dialects.test import Test
from xdsl.ir import Attribute, ParametrizedAttribute
from xdsl.irdl import (
    IRDLOperation,
    irdl_attr_definition,
    irdl_op_definition,
    prop_def,
    region_def,
)
from xdsl.parser import Parser
from xdsl.printer import Printer
from xdsl.utils.exceptions import ParseError
from xdsl.utils.mlir_lexer import (
    KIND_BY_PUNCTUATION_SPELLING,
    MLIRTokenKind,
    PunctuationSpelling,
)
from xdsl.utils.str_enum import StrEnum

# pyright: reportPrivateUsage=false


@pytest.mark.parametrize(
    "data",
    [
        dict(a=IntAttr(1), b=IntAttr(2), c=IntAttr(3)),
        dict(
            a=StringAttr("hello"),
            b=IntAttr(2),
            c=ArrayAttr([IntAttr(2), StringAttr("world")]),
        ),
        {},
    ],
)
def test_dictionary_attr(data: dict[str, Attribute]):
    attr = DictionaryAttr(data)

    with StringIO() as io:
        Printer(io).print_attribute(attr)
        text = io.getvalue()

    ctx = Context()
    ctx.load_dialect(Builtin)

    attr1 = Parser(ctx, text).parse_attribute()
    attr2 = Parser(ctx, "attributes " + text).parse_optional_attr_dict_with_keyword()
    assert isinstance(attr1, DictionaryAttr)
    assert isinstance(attr2, DictionaryAttr)

    assert attr1.data == data
    assert attr2.data == data


@pytest.mark.parametrize("text", ["{}", "{a = 1}", "attr {}"])
def test_dictionary_attr_with_keyword_missing(text: str):
    ctx = Context()
    ctx.load_dialect(Builtin)

    assert Parser(ctx, text).parse_optional_attr_dict_with_keyword() is None


@pytest.mark.parametrize(
    "text, reserved_names",
    [("{a = 1}", ["a"]), ("{a = 1}", ["b", "a"]), ("{b = 1, a = 1}", ["a"])],
)
def test_dictionary_attr_with_keyword_reserved(text: str, reserved_names: list[str]):
    ctx = Context()
    ctx.load_dialect(Builtin)

    with pytest.raises(ParseError):
        Parser(ctx, "attributes" + text).parse_optional_attr_dict_with_keyword(
            reserved_names
        )


@pytest.mark.parametrize(
    "text, expected",
    [("{b = 1}", ["a"]), ("{c = 1}", ["b", "a"]), ("{b = 1, a = 1}", ["c"])],
)
def test_dictionary_attr_with_keyword_not_reserved(text: str, expected: list[str]):
    ctx = Context()
    ctx.load_dialect(Builtin)

    res = Parser(ctx, "attributes" + text).parse_optional_attr_dict_with_keyword(
        expected
    )
    assert isinstance(res, DictionaryAttr)


@irdl_attr_definition
class DummyAttr(ParametrizedAttribute):
    name = "dummy.attr"


def test_parsing():
    """
    Test that the default attribute parser does not try to
    parse attribute arguments without the delimiters.
    """
    ctx = Context()
    ctx.load_attr_or_type(DummyAttr)

    prog = '#dummy.attr "foo"'
    parser = Parser(ctx, prog)

    r = parser.parse_attribute()
    assert r == DummyAttr()


@pytest.mark.parametrize(
    "text,expected",
    [
        ("@a", StringAttr("a")),
        ("@_", StringAttr("_")),
        ("@a1_2", StringAttr("a1_2")),
        ("@a$_.", StringAttr("a$_.")),
        ('@"foo"', StringAttr("foo")),
        ('@"@"', StringAttr("@")),
        ('@"\\t"', StringAttr("\t")),
        ("f", None),
        ('"f"', None),
    ],
)
def test_symbol_name(text: str, expected: StringAttr | None):
    ctx = Context()
    ctx.load_dialect(Builtin)

    parser = Parser(ctx, text)
    assert parser.parse_optional_symbol_name() == expected

    parser = Parser(ctx, text)
    if expected is not None:
        assert parser.parse_symbol_name() == expected
    else:
        with pytest.raises(ParseError):
            parser.parse_symbol_name()


@pytest.mark.parametrize(
    "ref,expected",
    [
        ("@foo", SymbolRefAttr("foo")),
        ("@foo::@bar", SymbolRefAttr("foo", ["bar"])),
        ("@foo::@bar:", SymbolRefAttr("foo", ["bar"])),
        ('@foo::@"bar"', SymbolRefAttr("foo", ["bar"])),
        ("@foo::@bar::@baz", SymbolRefAttr("foo", ["bar", "baz"])),
    ],
)
def test_symref(ref: str, expected: Attribute | None):
    """
    Test that symbol references are correctly parsed.
    """
    ctx = Context()
    ctx.load_dialect(Builtin)

    parser = Parser(ctx, ref)
    parsed_ref = parser.parse_attribute()

    assert parsed_ref == expected


@pytest.mark.parametrize(
    "text,expected,expect_type",
    [
        ("%foo", ("foo", None), False),
        ("%-1foo", ("-1foo", None), False),
        ("%foo : i32", ("foo", i32), True),
        ("i32 : %bar", None, False),
        ("i32 : %bar", None, True),
        ("i32 %bar", None, False),
        ("i32 %bar", None, True),
        ("i32", None, False),
        ("i32", None, True),
    ],
)
def test_parse_argument(
    text: str, expected: tuple[str, Attribute | None] | None, expect_type: bool
):
    """
    arg ::= percent-id              if expect_type is False
    arg ::= percent-id ':' type     if expect_type is True
    """
    ctx = Context()
    ctx.load_dialect(Builtin)

    # parse_optional_argument
    parser = Parser(ctx, text)
    res = parser.parse_optional_argument(expect_type)
    if expected is not None:
        assert res is not None
        assert res.name.text[1:] == expected[0]
        if expected[1] is not None:
            assert isinstance(res, Parser.Argument)
            assert res.type == expected[1]
        else:
            assert isinstance(res, parser.UnresolvedArgument)
    else:
        assert res is None

    # parse_argument
    parser = Parser(ctx, text)
    if expected is not None:
        res = parser.parse_argument(expect_type=expect_type)
        assert res is not None
        assert res.name.text[1:] == expected[0]
        if expected[1] is not None:
            assert isinstance(res, Parser.Argument)
            assert res.type == expected[1]
        else:
            assert isinstance(res, parser.UnresolvedArgument)
    else:
        with pytest.raises(ParseError):
            parser.parse_argument(expect_type=expect_type)


@pytest.mark.parametrize(
    "text,expect_type",
    [
        ("%foo : %bar", True),
        ("%foo : ", True),
        ("%foo", True),
        ("%foo %bar", True),
    ],
)
def test_parse_argument_fail(text: str, expect_type: bool):
    ctx = Context()
    ctx.load_dialect(Builtin)

    # parse_optional_argument
    parser = Parser(ctx, text)
    with pytest.raises(ParseError):
        parser.parse_optional_argument(expect_type)

    # parse_argument
    parser = Parser(ctx, text)
    with pytest.raises(ParseError):
        parser.parse_argument(expect_type=expect_type)


@pytest.mark.parametrize(
    "text,num_ops_and_args",
    [
        # no blocks
        ("{}", []),
        # One entry block
        ("""{ "test.op"() : () -> () }""", [(1, 0)]),
        # One entry block and another block
        (
            """{
          "test.op"() : () -> ()
        ^bb0(%x: i32):
          "test.op"() : () -> ()
          "test.op"() : () -> ()
        }""",
            [(1, 0), (2, 1)],
        ),
        # One labeled entry block and another block
        (
            """{
        ^bb0:
          "test.op"() : () -> ()
          "test.op"() : () -> ()
        ^bb1:
          "test.op"() : () -> ()
        }""",
            [(2, 0), (1, 0)],
        ),
        # One labeled entry block with args and another block
        (
            """{
        ^bb0(%x: i32, %y: i32):
          "test.op"() : () -> ()
          "test.op"() : () -> ()
        ^bb1(%z: i32):
          "test.op"() : () -> ()
        }""",
            [(2, 2), (1, 1)],
        ),
        # Not regions
        ("""^bb:""", None),
        ("""}""", None),
        (""""test.op"() : () -> ()""", None),
    ],
)
def test_parse_region_no_args(
    text: str, num_ops_and_args: list[tuple[int, int]] | None
):
    ctx = Context()
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Test)

    parser = Parser(ctx, text)
    if num_ops_and_args is None:
        with pytest.raises(ParseError):
            parser.parse_region()
    else:
        res = parser.parse_region()
        for block, (n_ops, n_args) in zip(res.blocks, num_ops_and_args):
            assert len(block.ops) == n_ops
            assert len(block.args) == n_args

    parser = Parser(ctx, text)
    res = parser.parse_optional_region()
    if num_ops_and_args is None:
        assert res is None
    else:
        assert res is not None
        for block, (n_ops, n_args) in zip(res.blocks, num_ops_and_args):
            assert len(block.ops) == n_ops
            assert len(block.args) == n_args


@pytest.mark.parametrize(
    "text",
    [
        """{""",
        """{ "test.op"() : () -> ()""",
    ],
)
def test_parse_region_fail(text: str):
    ctx = Context()
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Test)

    parser = Parser(ctx, text)
    with pytest.raises(ParseError):
        parser.parse_region()

    parser = Parser(ctx, text)
    with pytest.raises(ParseError):
        parser.parse_optional_region()


@pytest.mark.parametrize(
    "text",
    [
        """%x : i32 { "test.op"(%x) : (i32) -> () }""",
        """%x : i32 { "test.op"(%x) : (i32) -> () ^bb0: }""",
    ],
)
def test_parse_region_with_args(text: str):
    """Parse a region with args already provided."""
    ctx = Context()
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Test)

    parser = Parser(ctx, text)
    arg = parser.parse_argument()
    region = parser.parse_region((arg,))
    assert len(region.blocks[0].args) == 1

    parser = Parser(ctx, text)
    arg = parser.parse_argument()
    region = parser.parse_optional_region((arg,))
    assert region is not None
    assert len(region.blocks[0].args) == 1


@pytest.mark.parametrize(
    "text",
    [
        """%x : i32 { ^bb: "test.op"(%x) : (i32) -> () }""",
        """%x : i32 { ^bb(%y : i32): "test.op"(%x) : (i32) -> () }""",
        """%x : i32 { %x = "test.op"() : () -> (i32) }""",
    ],
)
def test_parse_region_with_args_fail(text: str):
    """Parse a region with args already provided."""
    ctx = Context()
    ctx.load_dialect(Builtin)
    ctx.load_dialect(Test)

    parser = Parser(ctx, text)
    arg = parser.parse_argument()
    with pytest.raises(ParseError):
        parser.parse_region((arg,))

    parser = Parser(ctx, text)
    arg = parser.parse_argument()
    with pytest.raises(ParseError):
        parser.parse_optional_region((arg,))


@irdl_op_definition
class MultiRegionOp(IRDLOperation):
    name = "test.multi_region"
    r1 = region_def()
    r2 = region_def()


def test_parse_multi_region_mlir():
    ctx = Context()
    ctx.load_op(MultiRegionOp)

    op_str = """
    "test.multi_region" () ({
    }, {
    }) : () -> ()
    """

    parser = Parser(ctx, op_str)

    op = parser.parse_op()

    assert len(op.regions) == 2


def test_parse_block_name():
    block_str = """
    ^bb0(%name: i32, %100: i32):
    """

    ctx = Context()
    parser = Parser(ctx, block_str)
    block = parser._parse_block()

    assert block.args[0].name_hint == "name"
    assert block.args[1].name_hint is None


@pytest.mark.parametrize(
    "delimiter,open_bracket,close_bracket",
    [
        (Parser.Delimiter.NONE, "", ""),
        (Parser.Delimiter.PAREN, "(", ")"),
        (Parser.Delimiter.SQUARE, "[", "]"),
        (Parser.Delimiter.BRACES, "{", "}"),
        (Parser.Delimiter.ANGLE, "<", ">"),
    ],
)
def test_parse_comma_separated_list(
    delimiter: Parser.Delimiter, open_bracket: str, close_bracket: str
):
    input = open_bracket + "2, 4, 5" + close_bracket

    parser = Parser(Context(), input)
    res = parser.parse_comma_separated_list(delimiter, parser.parse_integer, " in test")
    assert res == [2, 4, 5]

    parser = Parser(Context(), input)
    if delimiter is Parser.Delimiter.NONE:
        res = parser.parse_optional_undelimited_comma_separated_list(
            parser.parse_optional_integer, parser.parse_integer
        )
    else:
        res = parser.parse_optional_comma_separated_list(
            delimiter, parser.parse_integer, " in test"
        )
    assert res == [2, 4, 5]


@pytest.mark.parametrize(
    "delimiter,open_bracket,close_bracket",
    [
        (Parser.Delimiter.PAREN, "(", ")"),
        (Parser.Delimiter.SQUARE, "[", "]"),
        (Parser.Delimiter.BRACES, "{", "}"),
        (Parser.Delimiter.ANGLE, "<", ">"),
    ],
)
def test_parse_comma_separated_list_empty(
    delimiter: Parser.Delimiter, open_bracket: str, close_bracket: str
):
    input = open_bracket + close_bracket
    parser = Parser(Context(), input)
    res = parser.parse_comma_separated_list(delimiter, parser.parse_integer, " in test")
    assert res == []


def test_parse_comma_separated_list_none_delimiter_empty():
    parser = Parser(Context(), "o")
    with pytest.raises(ParseError):
        parser.parse_comma_separated_list(
            Parser.Delimiter.NONE, parser.parse_integer, " in test"
        )


def test_parse_comma_separated_list_none_delimiter_two_no_comma():
    """Test that a list without commas will only parse the first element."""
    parser = Parser(Context(), "1 2")
    res = parser.parse_comma_separated_list(
        Parser.Delimiter.NONE, parser.parse_integer, " in test"
    )
    assert res == [1]
    assert parser.parse_optional_integer() is not None

    parser = Parser(Context(), "1 2")
    parser.parse_optional_undelimited_comma_separated_list(
        parser.parse_optional_integer, parser.parse_integer
    )
    assert res == [1]
    assert parser.parse_optional_integer() is not None


@pytest.mark.parametrize(
    "delimiter",
    [
        (Parser.Delimiter.PAREN),
        (Parser.Delimiter.SQUARE),
        (Parser.Delimiter.BRACES),
        (Parser.Delimiter.ANGLE),
    ],
)
def test_parse_optional_comma_separated_list(delimiter: Parser.Delimiter):
    parser = Parser(Context(), "o")
    res = parser.parse_optional_comma_separated_list(delimiter, parser.parse_integer)
    assert res is None


def test_parse_optional_undelimited_comma_separated_list_empty():
    parser = Parser(Context(), "o")
    res = parser.parse_optional_undelimited_comma_separated_list(
        parser.parse_optional_integer, parser.parse_integer
    )
    assert res is None


@pytest.mark.parametrize(
    "delimiter,open_bracket,close_bracket",
    [
        (Parser.Delimiter.PAREN, "(", ")"),
        (Parser.Delimiter.SQUARE, "[", "]"),
        (Parser.Delimiter.BRACES, "{", "}"),
        (Parser.Delimiter.ANGLE, "<", ">"),
    ],
)
def test_parse_comma_separated_list_error_element(
    delimiter: Parser.Delimiter, open_bracket: str, close_bracket: str
):
    input = open_bracket + "o" + close_bracket
    parser = Parser(Context(), input)
    with pytest.raises(ParseError, match="Expected integer literal"):
        parser.parse_comma_separated_list(delimiter, parser.parse_integer, " in test")

    parser = Parser(Context(), input)
    with pytest.raises(ParseError, match="Expected integer literal"):
        parser.parse_optional_comma_separated_list(
            delimiter, parser.parse_integer, " in test"
        )


@pytest.mark.parametrize(
    "delimiter,open_bracket,close_bracket",
    [
        (Parser.Delimiter.PAREN, "(", ")"),
        (Parser.Delimiter.SQUARE, "[", "]"),
        (Parser.Delimiter.BRACES, "{", "}"),
        (Parser.Delimiter.ANGLE, "<", ">"),
    ],
)
def test_parse_comma_separated_list_error_delimiters(
    delimiter: Parser.Delimiter, open_bracket: str, close_bracket: str
):
    input = open_bracket + "2, 4 5"
    parser = Parser(Context(), input)
    with pytest.raises(
        ParseError, match=re.escape(f"'{close_bracket}' expected in test")
    ) as e:
        parser.parse_comma_separated_list(delimiter, parser.parse_integer, " in test")
    assert e.value.span.text == "5"

    parser = Parser(Context(), input)
    with pytest.raises(
        ParseError, match=re.escape(f"'{close_bracket}' expected in test")
    ) as e:
        parser.parse_optional_comma_separated_list(
            delimiter, parser.parse_integer, " in test"
        )
    assert e.value.span.text == "5"


@pytest.mark.parametrize("punctuation", list(KIND_BY_PUNCTUATION_SPELLING.values()))
def test_is_punctuation_true(punctuation: MLIRTokenKind):
    assert punctuation.is_punctuation()


@pytest.mark.parametrize(
    "punctuation",
    [MLIRTokenKind.BARE_IDENT, MLIRTokenKind.EOF, MLIRTokenKind.INTEGER_LIT],
)
def test_is_punctuation_false(punctuation: MLIRTokenKind):
    assert not punctuation.is_punctuation()


@pytest.mark.parametrize("punctuation", list(KIND_BY_PUNCTUATION_SPELLING.values()))
def test_is_spelling_of_punctuation_true(punctuation: MLIRTokenKind):
    value = cast(PunctuationSpelling, punctuation.value)
    assert MLIRTokenKind.is_spelling_of_punctuation(value)


@pytest.mark.parametrize("punctuation", [">-", "o", "4", "$", "_", "@"])
def test_is_spelling_of_punctuation_false(punctuation: str):
    assert not MLIRTokenKind.is_spelling_of_punctuation(punctuation)


@pytest.mark.parametrize("punctuation", list(KIND_BY_PUNCTUATION_SPELLING.values()))
def test_get_punctuation_kind(punctuation: MLIRTokenKind):
    value = cast(PunctuationSpelling, punctuation.value)
    assert punctuation.get_punctuation_kind_from_name(value) == punctuation


@pytest.mark.parametrize("punctuation", list(KIND_BY_PUNCTUATION_SPELLING.keys()))
def test_parse_punctuation(punctuation: PunctuationSpelling):
    parser = Parser(Context(), punctuation)

    res = parser.parse_punctuation(punctuation)
    assert res == punctuation
    assert parser._parse_token(MLIRTokenKind.EOF, "").kind == MLIRTokenKind.EOF


@pytest.mark.parametrize("punctuation", list(KIND_BY_PUNCTUATION_SPELLING.keys()))
def test_parse_punctuation_fail(punctuation: PunctuationSpelling):
    parser = Parser(Context(), "e +")
    with pytest.raises(ParseError) as e:
        parser.parse_punctuation(punctuation, " in test")
    assert e.value.span.text == "e"
    assert e.value.msg == "Expected '" + punctuation + "' in test"


@pytest.mark.parametrize("punctuation", list(KIND_BY_PUNCTUATION_SPELLING.keys()))
def test_parse_optional_punctuation(punctuation: PunctuationSpelling):
    parser = Parser(Context(), punctuation)
    res = parser.parse_optional_punctuation(punctuation)
    assert res == punctuation
    assert parser._parse_token(MLIRTokenKind.EOF, "").kind == MLIRTokenKind.EOF


@pytest.mark.parametrize("punctuation", list(KIND_BY_PUNCTUATION_SPELLING.keys()))
def test_parse_optional_punctuation_fail(punctuation: PunctuationSpelling):
    parser = Parser(Context(), "e +")
    assert parser.parse_optional_punctuation(punctuation) is None


@pytest.mark.parametrize(
    "text, expected_value",
    [
        ("true", True),
        ("false", False),
        ("True", None),
        ("False", None),
    ],
)
def test_parse_boolean(text: str, expected_value: bool | None):
    parser = Parser(Context(), text)
    assert parser.parse_optional_boolean() == expected_value

    parser = Parser(Context(), text)
    if expected_value is None:
        with pytest.raises(ParseError):
            parser.parse_boolean()
    else:
        assert parser.parse_boolean() == expected_value


@pytest.mark.parametrize(
    "text, expected_value, allow_boolean, allow_negative",
    [
        ("42", 42, False, False),
        ("42", 42, True, False),
        ("42", 42, False, True),
        ("42", 42, True, True),
        ("-1", None, False, False),
        ("-1", None, True, False),
        ("-1", -1, False, True),
        ("-1", -1, True, True),
        ("true", None, False, False),
        ("true", 1, True, False),
        ("true", None, False, True),
        ("true", 1, True, True),
        ("false", None, False, False),
        ("false", 0, True, False),
        ("false", None, False, True),
        ("false", 0, True, True),
        ("True", None, True, True),
        ("False", None, True, True),
        ("0x1a", 26, False, False),
        ("0x1a", 26, True, False),
        ("0x1a", 26, False, True),
        ("0x1a", 26, True, True),
        ("-0x1a", None, False, False),
        ("-0x1a", None, True, False),
        ("-0x1a", -26, False, True),
        ("-0x1a", -26, True, True),
    ],
)
def test_parse_int(
    text: str, expected_value: int | None, allow_boolean: bool, allow_negative: bool
):
    parser = Parser(Context(), text)
    assert (
        parser.parse_optional_integer(
            allow_boolean=allow_boolean, allow_negative=allow_negative
        )
        == expected_value
    )

    parser = Parser(Context(), text)
    if expected_value is None:
        with pytest.raises(ParseError):
            parser.parse_integer(
                allow_boolean=allow_boolean, allow_negative=allow_negative
            )
    else:
        assert (
            parser.parse_integer(
                allow_boolean=allow_boolean, allow_negative=allow_negative
            )
            == expected_value
        )


@pytest.mark.parametrize("nonnumeric", ["--", "+", "a", "{", "(1.0, 1.0)"])
def test_parse_optional_bool_int_or_float_nonnumeric(nonnumeric: str):
    parser = Parser(Context(), nonnumeric)
    assert parser._parse_optional_bool_int_or_float() is None


@pytest.mark.parametrize(
    "numeric,typ",
    [
        ("1", int),
        ("true", bool),
        ("1.0", float),
        ("-1", int),
        ("-1.0", float),
        ("false", bool),
    ],
)
def test_parse_optional_bool_int_or_float_numeric(numeric: str, typ: type):
    parser = Parser(Context(), numeric)
    value_span = parser._parse_optional_bool_int_or_float()
    assert value_span is not None
    value, span = value_span
    match typ:
        case builtins.int:
            assert value == typ(numeric)
            assert span.text == numeric
        case builtins.float:
            assert value == typ(numeric)
            assert span.text == numeric
        case builtins.bool:
            expected = numeric == "true"
            assert value == expected
            assert span.text == numeric
        case _:
            pytest.fail("unreachable")


@pytest.mark.parametrize("nonnumeric", ["--", "+", "a", "{", "(1.0, 1.0)"])
def test_parse_bool_int_or_float_nonnumeric(nonnumeric: str):
    parser = Parser(Context(), nonnumeric)
    with pytest.raises(ParseError):
        parser._parse_bool_int_or_float()


@pytest.mark.parametrize(
    "numeric,typ",
    [
        ("1", int),
        ("true", bool),
        ("1.0", float),
        ("-1", int),
        ("-1.0", float),
        ("false", bool),
    ],
)
def test_parse_bool_int_or_float_numeric(numeric: str, typ: type):
    parser = Parser(Context(), numeric)
    value_span = parser._parse_optional_bool_int_or_float()
    assert value_span is not None
    value, span = value_span
    match typ:
        case builtins.int:
            assert value == typ(numeric)
            assert span.text == numeric
        case builtins.float:
            assert value == typ(numeric)
            assert span.text == numeric
        case builtins.bool:
            expected = numeric == "true"
            assert value == expected
            assert span.text == numeric
        case _:
            pytest.fail("unreachable")


@pytest.mark.parametrize(
    "text, allow_boolean, allow_negative",
    [
        ("-false", False, True),
        ("-false", True, True),
        ("-true", False, True),
        ("-true", True, True),
        ("-k", True, True),
        ("-(", False, True),
    ],
)
def test_parse_optional_int_error(text: str, allow_boolean: bool, allow_negative: bool):
    """Test that parsing a negative without an integer after raise an error."""
    parser = Parser(Context(), text)
    with pytest.raises(ParseError):
        parser.parse_optional_integer(
            allow_boolean=allow_boolean, allow_negative=allow_negative
        )

    parser = Parser(Context(), text)
    with pytest.raises(ParseError):
        parser.parse_integer(allow_boolean=allow_boolean, allow_negative=allow_negative)


@pytest.mark.parametrize(
    "text, allow_boolean, expected_value",
    [
        ("42", False, 42),
        ("-1", False, -1),
        ("true", False, None),
        ("false", False, None),
        ("0x1a", False, 26),
        ("-0x1a", False, -26),
        ("0.", False, 0.0),
        ("1.", False, 1.0),
        ("0.2", False, 0.2),
        ("38.1243", False, 38.1243),
        ("92.54e43", False, 92.54e43),
        ("92.5E43", False, 92.5e43),
        ("43.3e-54", False, 43.3e-54),
        ("32.E+25", False, 32.0e25),
        ("true", True, 1),
        ("false", True, 0),
        ("42", True, 42),
        ("0.2", True, 0.2),
    ],
)
def test_parse_number(
    text: str, allow_boolean: bool, expected_value: int | float | None
):
    parser = Parser(Context(), text)
    assert parser.parse_optional_number(allow_boolean=allow_boolean) == expected_value

    parser = Parser(Context(), text)
    if expected_value is None:
        with pytest.raises(ParseError):
            parser.parse_number()
    else:
        assert parser.parse_number(allow_boolean=allow_boolean) == expected_value


@pytest.mark.parametrize(
    "text, expected_value",
    [
        ("3: i16", IntegerAttr(3, 16)),
        ("24: i32", IntegerAttr(24, 32)),
        ("0: index", IntegerAttr.from_index_int_value(0)),
        ("-64: i64", IntegerAttr(-64, 64)),
        ("-64.4: f64", FloatAttr(-64.4, 64)),
        ("32.4: f32", FloatAttr(32.4, 32)),
        ("0x7e00 : f16", FloatAttr(float("nan"), 16)),
        ("0x7c00 : f16", FloatAttr(float("inf"), 16)),
        ("0xfc00 : f16", FloatAttr(float("-inf"), 16)),
        ("0x7fc00000 : f32", FloatAttr(float("nan"), 32)),
        ("0x7f800000 : f32", FloatAttr(float("inf"), 32)),
        ("0xff800000 : f32", FloatAttr(float("-inf"), 32)),
        ("0x7ff8000000000000 : f64", FloatAttr(float("nan"), 64)),
        ("0x7ff0000000000000 : f64", FloatAttr(float("inf"), 64)),
        ("0xfff0000000000000 : f64", FloatAttr(float("-inf"), 64)),
        # ("3 : f64", None),  # todo this fails in mlir-opt but not in xdsl
    ],
)
def test_parse_optional_builtin_int_or_float_attr(
    text: str, expected_value: IntegerAttr | FloatAttr
):
    parser = Parser(Context(), text)
    assert parser.parse_optional_builtin_int_or_float_attr() == expected_value


@pytest.mark.parametrize(
    "text, allow_boolean",
    [
        ("-false", False),
        ("-true", False),
        ("-false", True),
        ("-true", True),
        ("-k", False),
        ("-(", False),
    ],
)
def test_parse_number_error(text: str, allow_boolean: bool):
    """
    Test that parsing a negative without an
    integer or a float after raise an error.
    """
    parser = Parser(Context(), text)
    with pytest.raises(ParseError):
        parser.parse_optional_number(allow_boolean=allow_boolean)

    parser = Parser(Context(), text)
    with pytest.raises(ParseError):
        parser.parse_number(allow_boolean=allow_boolean)


@irdl_op_definition
class PropertyOp(IRDLOperation):
    name = "test.prop_op"

    first = prop_def(StringAttr)
    second = prop_def(IntegerAttr[IntegerType])


def test_properties_retrocompatibility():
    # Straightforward case
    ctx = Context()
    ctx.load_op(PropertyOp)
    parser = Parser(ctx, '"test.prop_op"() <{first = "str", second = 42}> : () -> ()')

    op = parser.parse_op()
    assert isinstance(op, PropertyOp)
    op.verify()

    # Retrocompatibility case, only target
    parser = Parser(ctx, '"test.prop_op"() {first = "str", second = 42} : () -> ()')
    retro_op = parser.parse_op()
    assert isinstance(retro_op, PropertyOp)
    retro_op.verify()

    assert op.attributes == retro_op.attributes
    assert op.properties == retro_op.properties

    # Test partial case
    parser = Parser(ctx, '"test.prop_op"() <{first = "str"}> {second = 42} : () -> ()')
    partial_op = parser.parse_op()
    assert isinstance(partial_op, PropertyOp)
    partial_op.verify()

    assert op.attributes == partial_op.attributes
    assert op.properties == partial_op.properties


def test_parse_location():
    ctx = Context()
    attr = Parser(ctx, "loc(unknown)").parse_optional_location()
    assert attr == UnknownLoc()

    attr = Parser(ctx, 'loc("one":2:3)').parse_optional_location()
    assert attr == FileLineColLoc(StringAttr("one"), IntAttr(2), IntAttr(3))

    with pytest.raises(ParseError, match="Unexpected location syntax."):
        Parser(ctx, "loc(unexpected)").parse_optional_location()


@pytest.mark.parametrize(
    "keyword,expected",
    [
        ("public", StringAttr("public")),
        ("nested", StringAttr("nested")),
        ("private", StringAttr("private")),
        ("privateeee", None),
        ("unknown", None),
    ],
)
def test_parse_visibility(keyword: str, expected: StringAttr | None):
    assert Parser(Context(), keyword).parse_optional_visibility_keyword() == expected

    parser = Parser(Context(), keyword)
    if expected is None:
        with pytest.raises(ParseError, match="expect symbol visibility keyword"):
            parser.parse_visibility_keyword()
    else:
        assert parser.parse_visibility_keyword() == expected


class MyEnum(StrEnum):
    A = "a"
    B = "b"
    C = "c"
    D = "d-non-keyword"


@pytest.mark.parametrize(
    "keyword, expected",
    [
        ("a", MyEnum.A),
        ('"a"', MyEnum.A),
        ("b", MyEnum.B),
        ('"b"', MyEnum.B),
        ("c", MyEnum.C),
        ('"c"', MyEnum.C),
        ('"d-non-keyword"', MyEnum.D),
        ("other", None),
        ('"other"', None),
    ],
)
def test_parse_str_enum_right_token(keyword: str, expected: MyEnum | None):
    """
    Test parsing of string enums where the next
    token is a keyword or a string literal.
    """
    if expected is None:
        with pytest.raises(
            ParseError, match="Expected `a`, `b`, `c`, or `d-non-keyword`"
        ):
            Parser(Context(), keyword).parse_optional_str_enum(MyEnum)
    else:
        assert Parser(Context(), keyword).parse_optional_str_enum(MyEnum) == expected

    parser = Parser(Context(), keyword)
    if expected is None:
        with pytest.raises(
            ParseError, match="Expected `a`, `b`, `c`, or `d-non-keyword`"
        ):
            parser.parse_str_enum(MyEnum)
    else:
        assert parser.parse_str_enum(MyEnum) == expected


@pytest.mark.parametrize("keyword", ["2", "-"])
def test_parse_str_enum_wrong_token(keyword: str):
    """
    Test parsing of string enums where the next
    token is a keyword or a string literal.
    """
    assert Parser(Context(), keyword).parse_optional_str_enum(MyEnum) is None

    with pytest.raises(ParseError, match="Expected `a`, `b`, `c`, or `d-non-keyword`"):
        Parser(Context(), keyword).parse_str_enum(MyEnum)


@pytest.mark.parametrize("value", ["(1., 2)", "(1, 2.)"])
def test_parse_optional_complex_error(value: str):
    parser = Parser(Context(), value)
    with pytest.raises(
        ParseError,
        match=re.escape("Complex value must be either (float, float) or (int, int)"),
    ):
        parser._parse_optional_complex()


@pytest.mark.parametrize("noncomplex", ["1", "-1", "A", "{"])
def test_parse_optional_complex_noncomplex(noncomplex: str):
    parser = Parser(Context(), noncomplex)
    assert parser._parse_optional_complex() is None


@pytest.mark.parametrize(
    "toks, pyval", [("(-1., 2.)", (-1.0, 2.0)), ("(1, 2)", (1, 2))]
)
def test_parse_optional_complex_success(
    toks: str, pyval: tuple[int, int] | tuple[float, float]
):
    parser = Parser(Context(), toks)
    value_and_span = parser._parse_optional_complex()
    assert value_and_span is not None
    value, span = value_and_span
    assert value == pyval
    assert span.text == toks


@pytest.mark.parametrize(
    "start, end, text",
    [
        ("<", ">", "<1>"),
        ("[", "]", "[1]"),
        ("{", "}", "{1}"),
        ("(", ")", "(1)"),
        ("{-#", "#-}", "{-# 1 #-}"),
    ],
)
def test_delimiters(start: str, end: str, text: str):
    parser = Parser(Context(), text)
    with parser.delimited(start, end):
        value = parser.parse_integer()

    assert value == 1
    assert parser.pos == len(text)


def test_angle_brackets():
    parser = Parser(Context(), "<1>")
    with parser.in_angle_brackets():
        value = parser.parse_integer()

    assert value == 1
    assert parser.pos == 3


def test_square_brackets():
    parser = Parser(Context(), "[1]")
    with parser.in_square_brackets():
        value = parser.parse_integer()

    assert value == 1
    assert parser.pos == 3


def test_braces():
    parser = Parser(Context(), "{1}")
    with parser.in_braces():
        value = parser.parse_integer()

    assert value == 1
    assert parser.pos == 3


def test_parens():
    parser = Parser(Context(), "(1)")
    with parser.in_parens():
        value = parser.parse_integer()

    assert value == 1
    assert parser.pos == 3


def test_wrong_delimiter():
    parser = Parser(Context(), "(1)")
    with pytest.raises(ParseError, match="'<' expected"):
        with parser.in_angle_brackets():
            parser.parse_integer()


def test_early_return_delimiter():
    def my_parse(parser: Parser):
        with parser.in_angle_brackets():
            if parser.parse_integer() == 1:
                return 1
            return 2

    parser = Parser(Context(), "<1>")
    assert my_parse(parser) == 1
    assert parser.pos == 3


class MySingletonEnum(StrEnum):
    A = "a"


def test_parse_singleton_enum_fail():
    parser = Parser(Context(), "b")
    with pytest.raises(ParseError, match="Expected `a`"):
        parser.parse_str_enum(MySingletonEnum)


def test_metadata_parsing():
    ctx = Context()
    ctx.register_dialect("test", lambda: Test)
    metadata_dict = '{-# dialect_resources: {test: {some_res: "0x1"}} #-}'

    parser = Parser(ctx, metadata_dict)
    assert parser._parse_file_metadata_dictionary() is None

    test_dialect = ctx.get_dialect("test")
    interface = test_dialect.get_interface(OpAsmDialectInterface)
    assert interface

    element = interface.lookup("some_res")
    assert element == "0x1"


@pytest.mark.parametrize(
    "input, expected",
    [
        ("a", "a"),
        ('"a"', "a"),
        ("a-b", "a"),
        ('"a-b"', "a-b"),
        ("2a", None),
    ],
)
def test_parse_identifier_or_str_literal(input: str, expected: str | None):
    parser = Parser(Context(), input)
    result = parser.parse_optional_identifier_or_str_literal()
    assert result == expected

    parser = Parser(Context(), input)
    if expected is None:
        with pytest.raises(ParseError, match="identifier or string literal expected"):
            parser.parse_identifier_or_str_literal()
    else:
        assert parser.parse_identifier_or_str_literal() == expected



"""
Utilities used by reshape ops.
See [MLIR counterpart](https://github.com/llvm/llvm-project/blob/main/mlir/include/mlir/Dialect/Utils/ReshapeOpsUtils.h)
for more details.
"""

from dataclasses import dataclass
from typing import cast

from typing_extensions import TypeVar

from xdsl.dialects.builtin import I64, Annotated, ArrayAttr, IntegerAttr, ShapedType
from xdsl.ir import Attribute
from xdsl.irdl import (
    AtLeast,
    AttrConstraint,
    ConstraintContext,
    irdl_to_attr_constraint,
)
from xdsl.utils.exceptions import VerifyException

_CONTIGUOUS_ARRAY_TYPE_CONSTRAINT = irdl_to_attr_constraint(
    ArrayAttr[
        ArrayAttr[
            Annotated[
                IntegerAttr[I64],
                IntegerAttr.constr(value=AtLeast(0)),
            ]
        ]
    ]
)

ArrayOfIntArrayAttr = ArrayAttr[ArrayAttr[IntegerAttr]]


@dataclass(frozen=True)
class ContiguousArrayOfIntArray(AttrConstraint[ArrayOfIntArrayAttr]):
    """
    Enforce an ArrayAttr of ArrayAttr[IntegerAttr] to contain contiguous integer values across all inner arrays.
    For example: [[0, 1], [2, 3]] is valid, but [[3, 4], [0, 1]] is not.
    An empty inner array is considered contiguous.
    """

    def verify(self, attr: Attribute, constraint_context: ConstraintContext) -> None:
        _CONTIGUOUS_ARRAY_TYPE_CONSTRAINT.verify(
            attr, constraint_context=constraint_context
        )
        attr = cast(ArrayOfIntArrayAttr, attr)

        # Flatten all integer values from all inner arrays
        flat_values = [e.value.data for inner in attr.data for e in inner.data]
        # Check that the flattened list is contiguous
        for prev, curr in zip(flat_values, flat_values[1:]):
            if curr != prev + 1:
                raise VerifyException(f"All inner arrays must be contiguous: {attr}")

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> "ContiguousArrayOfIntArray":
        # No type variables to map in this constraint
        return self


def verify_reshape_like_types(
    collapsed_type: ShapedType,
    expanded_type: ShapedType,
    reassociation: ArrayAttr[ArrayAttr[IntegerAttr]],
):
    """
    Verify that collapsed and expanded types conform to reassociation mapping.
    """
    expanded_rank = len(expanded_type.get_shape())
    collapsed_rank = len(collapsed_type.get_shape())

    if expanded_rank < collapsed_rank:
        raise VerifyException(
            f"expected the expanded type, {expanded_type} to have a higher (or same) rank "
            f"than the collapsed type, {collapsed_type}."
        )

    if collapsed_rank != len(reassociation):
        raise VerifyException(
            f"expected collapsed rank ({collapsed_rank}) to equal the number of "
            f"reassociation maps ({len(reassociation)})."
        )

    # Check that the total reassociation dimensions match the expanded type's rank.
    total_reassociation_dims = sum(len(rm) for rm in reassociation)
    if total_reassociation_dims != expanded_rank:
        raise VerifyException(
            f"expected the total number of reassociation dimensions ({total_reassociation_dims}) "
            f"to equal the expanded type's rank ({expanded_rank})."
        )

    verify_reshape_like_shapes_are_compatible(
        collapsed_shape=collapsed_type.get_shape(),
        expanded_shape=expanded_type.get_shape(),
        reassociation=reassociation,
    )


def verify_reshape_like_shapes_are_compatible(
    collapsed_shape: tuple[int, ...],
    expanded_shape: tuple[int, ...],
    reassociation: ArrayOfIntArrayAttr,
):
    """
    Verify that collapsed and expanded shapes adhere to reassociation mapping.
    """
    expanded_dim_start = 0

    for map_idx, rm in enumerate(reassociation):
        found_dynamic = False
        linearized_static = 1

        # Look at the next `len(rm)` dims in expanded_shape
        for dim in expanded_shape[expanded_dim_start : expanded_dim_start + len(rm)]:
            if dim == -1:
                found_dynamic = True
            else:
                linearized_static *= dim

        if found_dynamic:
            # if any is dynamic, the collapsed must be dynamic too
            if not collapsed_shape[map_idx] == -1:
                raise VerifyException(
                    f"expected dimension {map_idx} of collapsed type to be dynamic "
                    f"since one or more of the corresponding dimensions in the "
                    f"expanded type is dynamic"
                )
        else:
            # all static → product must match
            if collapsed_shape[map_idx] != linearized_static:
                raise VerifyException(
                    f"expected dimension {map_idx} of collapsed type to be static "
                    f"value of {linearized_static}"
                )

        expanded_dim_start += len(rm)
